r"""

.. autofunction:: openpnm.models.physics.flow_shape_factors.ball_and_stick
.. autofunction:: openpnm.models.physics.flow_shape_factors.conical_frustum_and_stick

"""
from scipy import pi as _pi
from numpy import arctanh as _atanh
import numpy as _np
import scipy as _sp


def ball_and_stick(target, pore_area='pore.area',
                   throat_area='throat.area',
                   pore_diameter='pore.diameter',
                   throat_diameter='throat.diameter',
                   conduit_lengths='throat.conduit_lengths'):
    r"""
    Calculate conduit shape factors for hydraulic conductance, assuming
    pores and throats are spheres (balls) and constant cross-section
    cylinders (sticks).

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_area : string
        Dictionary key of the pore area values

    throat_area : string
        Dictionary key of the throat area values

    pore_diameter : string
        Dictionary key of the pore diameter values

    throat_diameter : string
        Dictionary key of the throat diameter values

    conduit_lengths : string
        Dictionary key of the conduit lengths' values

    Returns
    -------
    SF : dictionary
        Dictionary containing conduit shape factors to be used in hagen-
        poiseuille hydraulic conductance model. Shape factors are accessible
        via the keys: 'pore1', 'pore2' and 'throat'.

    Notes
    -----
    (1) This model accounts for the variable cross-section area in spheres.

    (2) WARNING: This model could break if `conduit_lengths` does not
    correspond to an actual ball and stick! Example: pore length is greater
    than pore radius --> :(

    References
    ----------
    Akbari, M., Sinton, D., & Bahrami, M. (2011). Viscous flow in variable
    cross-section microchannels of arbitrary shapes. International Journal of
    Heat and Mass Transfer, 54(17-18), 3970-3978.

    """
    _np.warnings.filterwarnings('ignore', category=RuntimeWarning)
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    cn = network['throat.conns'][throats]
    # Get pore diameter
    D1 = network[pore_diameter][cn[:, 0]]
    D2 = network[pore_diameter][cn[:, 1]]
    Dt = network[throat_diameter][throats]
    # Get conduit lengths
    L1 = network[conduit_lengths + '.pore1'][throats]
    L2 = network[conduit_lengths + '.pore2'][throats]
    Lt = network[conduit_lengths + '.throat'][throats]
    # Get pore/throat baseline areas (the one used in generic conductance)
    A1 = network[pore_area][cn[:, 0]]
    A2 = network[pore_area][cn[:, 1]]
    At = network[throat_area][throats]
    # Preallocating F, SF
    # F is INTEGRAL(1/A^2) dx , x : 0 --> L
    F1, F2, Ft = _np.zeros((3, len(Lt)))
    SF1, SF2, SFt = _np.ones((3, len(Lt)))
    # Setting SF to 1 when Li = 0 (ex. boundary pores)
    # INFO: This is needed since area could also be zero, which confuses NumPy
    m1, m2, mt = [Li != 0 for Li in [L1, L2, Lt]]
    SF1[~m1] = SF2[~m2] = SFt[~mt] = 1
    # Handle the case where Dt >= Dp
    M1, M2 = [(Di <= Dt) & mi for Di, mi in zip([D1, D2], [m1, m2])]
    F1[M1] = 16/3 * (L1*(D1**2 + D1*Dt + Dt**2) / (D1**3 * Dt**3 * _pi**2))[M1]
    F2[M2] = 16/3 * (L2*(D2**2 + D2*Dt + Dt**2) / (D2**3 * Dt**3 * _pi**2))[M2]
    # Handle the rest (true balls and sticks)
    N1, N2 = [(Di > Dt) & mi for Di, mi in zip([D1, D2], [m1, m2])]
    F1[N1] = (4/(D1**3*_pi**2) * ((2*D1*L1) / (D1**2-4*L1**2) + _atanh(2*L1/D1)))[N1]
    F2[N2] = (4/(D2**3*_pi**2) * ((2*D2*L2) / (D2**2-4*L2**2) + _atanh(2*L2/D2)))[N2]
    Ft[mt] = (Lt / At**2)[mt]
    # Calculate conduit shape factors
    SF1[m1] = (L1 / (A1**2 * F1))[m1]
    SF2[m2] = (L2 / (A2**2 * F2))[m2]
    SFt[mt] = (Lt / (At**2 * Ft))[mt]
    _np.warnings.filterwarnings('default', category=RuntimeWarning)
    return {'pore1': SF1, 'throat': SFt, 'pore2': SF2}


def conical_frustum_and_stick(target, pore_area='pore.area',
                              throat_area='throat.area',
                              pore_diameter='pore.diameter',
                              throat_diameter='throat.diameter',
                              conduit_lengths='throat.conduit_lengths'):
    r"""
    Calculate conduit shape factors for hydraulic conductance, assuming
    pores and throats are truncated pyramids (frustums) and constant
    cross-section cylinders (sticks).

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_area : string
        Dictionary key of the pore area values

    throat_area : string
        Dictionary key of the throat area values

    pore_diameter : string
        Dictionary key of the pore diameter values

    throat_diameter : string
        Dictionary key of the throat diameter values

    conduit_lengths : string
        Dictionary key of the conduit lengths' values

    Returns
    -------
    SF : dictionary
        Dictionary containing conduit shape factors to be used in hagen-
        poiseuille hydraulic conductance model. Shape factors are accessible
        via the keys: 'pore1', 'pore2' and 'throat'.

    References
    ----------
    Akbari, M., Sinton, D., & Bahrami, M. (2011). Viscous flow in variable
    cross-section microchannels of arbitrary shapes. International Journal of
    Heat and Mass Transfer, 54(17-18), 3970-3978.

    """
    _np.warnings.filterwarnings('ignore', category=RuntimeWarning)
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    cn = network['throat.conns'][throats]
    # Get pore diameter
    D1 = network[pore_diameter][cn[:, 0]]
    D2 = network[pore_diameter][cn[:, 1]]
    Dt = network[throat_diameter][throats]
    # Get pore/throat baseline areas (the one used in generic conductance)
    A1 = network[pore_area][cn[:, 0]]
    A2 = network[pore_area][cn[:, 1]]
    At = network[throat_area][throats]
    # Get conduit lengths
    L1 = network[conduit_lengths + '.pore1'][throats]
    L2 = network[conduit_lengths + '.pore2'][throats]
    Lt = network[conduit_lengths + '.throat'][throats]
    # Preallocating F, SF
    # F is INTEGRAL(1/A^2) dx , x : 0 --> L
    F1, F2, Ft = _np.zeros((3, len(Lt)))
    SF1, SF2, SFt = _np.ones((3, len(Lt)))
    # Setting SF to 1 when Li = 0 (ex. boundary pores)
    # INFO: This is needed since area could also be zero, which confuses NumPy
    m1, m2, mt = [Li != 0 for Li in [L1, L2, Lt]]
    SF1[~m1] = SF2[~m2] = SFt[~mt] = 1
    # Calculate integral of 1/A^2
    F1[m1] = 16/3 * (L1*(D1**2 + D1*Dt + Dt**2) / (D1**3 * Dt**3 * _pi**2))[m1]
    F2[m2] = 16/3 * (L2*(D2**2 + D2*Dt + Dt**2) / (D2**3 * Dt**3 * _pi**2))[m2]
    Ft[mt] = (Lt/At**2)[mt]
    # Calculate conduit shape factors
    SF1[m1] = (L1 / (A1**2 * F1))[m1]
    SF2[m2] = (L2 / (A2**2 * F2))[m2]
    SFt[mt] = (Lt / (At**2 * Ft))[mt]
    _np.warnings.filterwarnings('default', category=RuntimeWarning)
    return {'pore1': SF1, 'throat': SFt, 'pore2': SF2}


def ball_and_stick_2D(target, pore_area='pore.area',
                      throat_area='throat.area',
                      pore_diameter='pore.diameter',
                      throat_diameter='throat.diameter',
                      conduit_lengths='throat.conduit_lengths'):
    r"""
    Calculate conduit shape factors for hydraulic conductance, assuming
    pores and throats are circles (balls) and rectangles (sticks).

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_area : string
        Dictionary key of the pore area values

    throat_area : string
        Dictionary key of the throat area values

    pore_diameter : string
        Dictionary key of the pore diameter values

    throat_diameter : string
        Dictionary key of the throat diameter values

    conduit_lengths : string
        Dictionary key of the conduit lengths' values

    Returns
    -------
    SF : dictionary
        Dictionary containing conduit shape factors to be used in hagen-
        poiseuille hydraulic conductance model. Shape factors are accessible
        via the keys: 'pore1', 'pore2' and 'throat'.

    Notes
    -----
    (1) This model accounts for the variable cross-section area in spheres.

    (2) WARNING: This model could break if `conduit_lengths` does not
    correspond to an actual ball and stick! Example: pore length is greater
    than pore radius --> :(

    References
    ----------
    Akbari, M., Sinton, D., & Bahrami, M. (2011). Viscous flow in variable
    cross-section microchannels of arbitrary shapes. International Journal of
    Heat and Mass Transfer, 54(17-18), 3970-3978.

    """
    _np.warnings.filterwarnings('ignore', category=RuntimeWarning)
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    cn = network['throat.conns'][throats]
    # Get pore diameter
    D1 = network[pore_diameter][cn[:, 0]]
    D2 = network[pore_diameter][cn[:, 1]]
    # Get conduit lengths
    L1 = network[conduit_lengths + '.pore1'][throats]
    L2 = network[conduit_lengths + '.pore2'][throats]
    Lt = network[conduit_lengths + '.throat'][throats]
    # Get pore/throat baseline areas (the one used in generic conductance)
    A1 = network[pore_area][cn[:, 0]]
    A2 = network[pore_area][cn[:, 1]]
    At = network[throat_area][throats]
    # Preallocating F, SF
    # F is INTEGRAL(1/A^2) dx , x : 0 --> L
    F1, F2, Ft = _np.zeros((3, len(Lt)))
    SF1, SF2, SFt = _np.ones((3, len(Lt)))
    # Setting SF to 1 when Li = 0 (ex. boundary pores)
    # INFO: This is needed since area could also be zero, which confuses NumPy
    m1, m2, mt = [Li != 0 for Li in [L1, L2, Lt]]
    SF1[~m1] = SF2[~m2] = SFt[~mt] = 1
    F1[m1] = (_atanh(2*L1/D1) / (2*D1))[m1]
    F2[m2] = (_atanh(2*L2/D2) / (2*D2))[m2]
    Ft[mt] = (Lt / At**2)[mt]
    # Calculate conduit shape factors
    SF1[m1] = (L1 / (A1**2 * F1))[m1]
    SF2[m2] = (L2 / (A2**2 * F2))[m2]
    SFt[mt] = (Lt / (At**2 * Ft))[mt]
    _np.warnings.filterwarnings('default', category=RuntimeWarning)
    return {'pore1': SF1, 'throat': SFt, 'pore2': SF2}



from openpnm.utils import logging, Project
from openpnm.network import Cubic
from openpnm.geometry import GenericGeometry
import openpnm.models.geometry as gm
logger = logging.getLogger(__name__)


class BereaCubic(Project):
    r"""
    A traditional Berea Sandstone on a Cubic lattice

    Berea Sandstone is one of the standard materials used on geoscience
    studies due to it's importance in oil reservoir engineering as well as
    having well defined pore structure.  This class creates a Cubic Network
    with the appropriate lattice spacing and connectivity, then adds a Geometry
    object with the necessary pore-scale models and prescribed parameters.

    Parameters
    ----------
    shape : array_like
        The number of pores along each direction of the domain.  All other
        aspects of this model are prescribed by the code.

    name : string, optional
        The name to give the Project

    Notes
    -----
    The source code for this Material is relatively straight-forward, so is a
    good example starting point for creating custom materials.

    References
    ----------
    [1] ???

    Examples
    --------

    """

    def __init__(self, shape, name=None, **kwargs):
        super().__init__(name=name)

        net = Cubic(shape=shape, spacing=1e-4, connectivity=26,
                    project=self, **kwargs)

        geom = GenericGeometry(network=net, pores=net.Ps, throats=net.Ts)

        geom.add_model(propname='pore.seed',
                       model=gm.pore_seed.spatially_correlated,
                       weights=[1, 1, 1])
        geom.add_model(propname='pore.diameter',
                       model=gm.pore_size.weibull,
                       shape=2.2, scale=2e-5, loc=1e-6)
        geom.add_model(propname='pore.volume',
                       model=gm.pore_volume.sphere)
        geom.add_model(propname='throat.length',
                       model=gm.throat_length.piecewise)
        geom.add_model(propname='throat.diameter',
                       model=gm.throat_size.from_neighbor_pores,
                       pore_prop='pore.diameter', mode='min')
        geom.add_model(propname='throat.volume',
                       model=gm.throat_volume.cylinder)

        geom.regenerate_models()



r"""

.. autofunction:: openpnm.models.phases.viscosity.water
.. autofunction:: openpnm.models.phases.viscosity.reynolds
.. autofunction:: openpnm.models.phases.viscosity.chung

"""
import numpy as np


def water(target, temperature='pore.temperature', salinity='pore.salinity'):
    r"""
    Calculates viscosity of pure water or seawater at atmospheric pressure
    using Eq. (22) given by Sharqawy et. al [1]. Values at temperature higher
    than the normal boiling temperature are calculated at the saturation
    pressure.

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated. This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    temperature : string
        The dictionary key containing the temperature values. Temperature must
        be in Kelvin for this emperical equation to work. Can be either a pore
        or throat array.

    salinity : string
        The dictionary key containing the salinity values. Salinity must be
        expressed in g of salt per kg of solution (ppt). Can be either a
        pore or throat array, but must be consistent with ``temperature``.

    Returns
    -------
    value : NumPy ndarray
        Array containing viscosity of water/seawater in [kg/m.s]

    Notes
    -----
     T must be in K, and S in g of salt per kg of phase, or ppt (parts per
        thousand)
    VALIDITY: 273 < T < 453 K; 0 < S < 150 g/kg;
    ACCURACY: 1.5 %

    References
    ----------
    [1] Sharqawy M. H., Lienhard J. H., and Zubair, S. M., Desalination and
        Water Treatment, 2010.

    """
    T = target[temperature]
    if salinity in target.keys():
        S = target[salinity]
    else:
        S = 0
    TC = T-273.15
    S = S/1000
    a1 = 1.5700386464E-01
    a2 = 6.4992620050E+01
    a3 = -9.1296496657E+01
    a4 = 4.2844324477E-05
    mu_w = a4 + 1/(a1*(TC+a2)**2+a3)
    a5 = 1.5409136040E+00
    a6 = 1.9981117208E-02
    a7 = -9.5203865864E-05
    a8 = 7.9739318223E+00
    a9 = -7.5614568881E-02
    a10 = 4.7237011074E-04
    A = a5 + a6*T + a7*T**2
    B = a8 + a9*T + a10*T**2
    mu_sw = mu_w*(1 + A*S + B*S**2)
    value = mu_sw
    return value


def reynolds(target, u0, b, temperature='pore.temperature'):
    r"""
    Uses exponential model by Reynolds [1] for the temperature dependance of
    shear viscosity

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    u0, b : float, array_like
            Coefficients of the viscosity exponential model (mu = u0*Exp(-b*T)
            where T is the temperature in Kelvin

    temperature : string
        The dictionary key containing the temperature values (K).  Can be
        either a pore or throat array.

    Returns
    -------
    value : NumPy ndarray
        Array containing viscosity values based on Reynolds model.

    [1] Reynolds O. (1886). Phil Trans Royal Soc London, v. 177, p.157.

    """
    value = u0*np.exp(b*target[temperature])
    return value


def chung(target, temperature='pore.temperature',
          mol_weight='pore.molecular_weight',
          critical_temperature='pore.critical_temperature',
          critical_volume='pore.critical_volume'):
    r"""
    Uses Chung et al. [1] model to estimate viscosity for gases at low
    pressure (much less than the critical pressure) at conditions of interest.

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    temperatre: string
        The dictionary key containing the temperature values (K)

    critical_temperature : string
        The dictionary key containing the temperature values (K)

    mol_weight: string
        The dictionary key containing the molecular weight values (kg/mol)

    critical_volume : string
        The dictionary key containing the critical volume values (m3/kmol)

    Returns
    -------
    value : NumPy ndarray
        Array containing viscosity values based on Chung model [kg/m.s].

    References
    ----------
    [1] Chung, T.H., Lee, L.L., and Starling, K.E., Applications of Kinetic Gas
        Theories and Multiparameter Correlation for Prediction of Dilute Gas
        Viscosity and Thermal Conductivity”, Ind. Eng. Chem. Fundam.23:8, 1984.

    """
    T = target[temperature]
    MW = target[mol_weight]
    Tc = target[critical_temperature]
    Vc = target[critical_volume]
    Tr = T / Tc
    Tstar = 1.2593*Tr
    A = 1.161415
    B = 0.14874
    C = 0.52487
    D = 0.77320
    E = 2.16178
    F = 2.43787
    omega = (A*(Tstar)**(-B)) + C*(np.exp(-D*Tstar)) + E*(np.exp(-F*Tstar))
    sigma = 0.809*(Vc**(1/3))
    value = 26.69E-9*np.sqrt(MW*T)/(omega*sigma**2)
    return value



r"""

.. autofunction:: openpnm.models.physics.ionic_conductance.poisson
.. autofunction:: openpnm.models.physics.ionic_conductance.laplace
.. autofunction:: openpnm.models.physics.ionic_conductance.electroneutrality

"""
import numpy as _np
import scipy as _sp
from openpnm.utils import logging
logger = logging.getLogger(__name__)


def poisson(target,
            pore_area='pore.area',
            throat_area='throat.area',
            pore_diffusivity='pore.diffusivity',
            throat_diffusivity='throat.diffusivity',
            conduit_lengths='throat.conduit_lengths',
            conduit_shape_factors='throat.poisson_shape_factors',
            pore_volume='pore.volume',
            pore_temperature='pore.temperature',
            throat_temperature='throat.temperature',
            pore_valence='pore.valence',
            throat_valence='throat.valence',
            pore_concentration='pore.concentration'):
    r"""
    Calculate the ionic conductance of conduits in network (using the Poisson
    equation for charge conservation), where a conduit is
    ( 1/2 pore - full throat - 1/2 pore ). See the notes section.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_area : string
        Dictionary key of the pore area values

    throat_area : string
        Dictionary key of the throat area values

    pore_diffusivity : string
        Dictionary key of the pore diffusivity values

    throat_diffusivity : string
        Dictionary key of the throat diffusivity values

    conduit_lengths : string
        Dictionary key of the conduit length values

    conduit_shape_factors : string
        Dictionary key of the conduit DIFFUSION shape factor values

    pore_volume : string
        Dictionary key of the pore volume values

    pore_temperature : string
        Dictionary key of the pore temperature values

    throat_temperature : string
        Dictionary key of the throat temperature values

    pore_valence : string
       Dictionary key of the pore ionic species valence values

    throat_valence : string
       Dictionary key of the throat ionic species valence values

    pore_concentration : string
       Dictionary key of the pore ionic species concentration values

    Returns
    -------
    g : ndarray
        Array containing ionic conductance values for conduits in the
        geometry attached to the given physics object.

    Notes
    -----
    (1) This function requires that all the necessary phase properties already
    be calculated.

    (2) This function calculates the specified property for the *entire*
    network then extracts the values for the appropriate throats at the end.

    (3) This function assumes cylindrical throats with constant cross-section
    area. Corrections for different shapes and variable cross-section area can
    be imposed by passing the proper conduit_shape_factors argument.

    (4) shape_factor depends on the physics of the problem, i.e. diffusion-like
    processes and fluid flow need different shape factors.

    """
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    phase = target.project.find_phase(target)
    cn = network['throat.conns'][throats]
    # Getting equivalent areas
    A1 = network[pore_area][cn[:, 0]]
    At = network[throat_area][throats]
    A2 = network[pore_area][cn[:, 1]]
    # Getting conduit lengths
    L1 = network[conduit_lengths + '.pore1'][throats]
    Lt = network[conduit_lengths + '.throat'][throats]
    L2 = network[conduit_lengths + '.pore2'][throats]
    # Preallocating g
    g1, g2, gt = _np.zeros((3, len(Lt)))
    # Setting g to inf when Li = 0 (ex. boundary pores)
    # INFO: This is needed since area could also be zero, which confuses NumPy
    m1, m2, mt = [Li != 0 for Li in [L1, L2, Lt]]
    g1[~m1] = g2[~m2] = gt[~mt] = _sp.inf
    # Getting shape factors
    try:
        SF1 = phase[conduit_shape_factors+'.pore1'][throats]
        SFt = phase[conduit_shape_factors+'.throat'][throats]
        SF2 = phase[conduit_shape_factors+'.pore2'][throats]
    except KeyError:
        SF1 = SF2 = SFt = 1.0
    # Poisson or Laplace
    epsilon0 = 8.854187817e-12
    epsilonr = phase['pore.permittivity'][0]
    g1[m1] = epsilon0 * epsilonr * (A1)[m1] / L1[m1]
    g2[m2] = epsilon0 * epsilonr * (A2)[m2] / L2[m2]
    gt[mt] = epsilon0 * epsilonr * (At)[mt] / Lt[mt]
    # Preallocating g_inv
    g_inv1, g_inv2, g_invt = _np.zeros((3, len(Lt)))
    f1, f2, ft = [gi != 0 for gi in [g1, g2, gt]]
    g_inv1[~f1] = g_inv2[~f2] = g_invt[~ft] = _sp.inf
    g_inv1[f1] = 1/g1[f1]
    g_inv2[f2] = 1/g2[f2]
    g_invt[ft] = 1/gt[ft]
    # Apply shape factors and calculate the final conductance
    g = g_inv1/SF1 + g_inv2/SF2 + g_invt/SFt
    g[g != 0] = g[g != 0]**(-1)
    return g


def poisson_2D(target,
               pore_diameter='pore.diameter',
               throat_diameter='throat.diameter',
               pore_diffusivity='pore.diffusivity',
               throat_diffusivity='throat.diffusivity',
               conduit_lengths='throat.conduit_lengths',
               conduit_shape_factors='throat.poisson_shape_factors',
               pore_volume='pore.volume',
               pore_temperature='pore.temperature',
               throat_temperature='throat.temperature',
               pore_valence='pore.valence',
               throat_valence='throat.valence',
               pore_concentration='pore.concentration'):
    r"""
    Calculate the ionic conductance of conduits in network (using the Poisson
    equation for charge conservation), where a conduit is
    ( 1/2 pore - full throat - 1/2 pore ). See the notes section.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_diameter : string
        Dictionary key of the pore diameter values

    throat_diameter : string
        Dictionary key of the throat diameter values

    pore_diffusivity : string
        Dictionary key of the pore diffusivity values

    throat_diffusivity : string
        Dictionary key of the throat diffusivity values

    conduit_lengths : string
        Dictionary key of the conduit length values

    conduit_shape_factors : string
        Dictionary key of the conduit DIFFUSION shape factor values

    pore_volume : string
        Dictionary key of the pore volume values

    pore_temperature : string
        Dictionary key of the pore temperature values

    throat_temperature : string
        Dictionary key of the throat temperature values

    pore_valence : string
       Dictionary key of the pore ionic species valence values

    throat_valence : string
       Dictionary key of the throat ionic species valence values

    pore_concentration : string
       Dictionary key of the pore ionic species concentration values

    Returns
    -------
    g : ndarray
        Array containing ionic conductance values for conduits in the
        geometry attached to the given physics object.

    Notes
    -----
    (1) This function requires that all the necessary phase properties already
    be calculated.

    (2) This function calculates the specified property for the *entire*
    network then extracts the values for the appropriate throats at the end.

    (3) This function assumes cylindrical throats with constant cross-section
    area. Corrections for different shapes and variable cross-section area can
    be imposed by passing the proper conduit_shape_factors argument.

    (4) shape_factor depends on the physics of the problem, i.e. diffusion-like
    processes and fluid flow need different shape factors.

    """
    return poisson(target=target,
                   pore_area=pore_diameter,
                   throat_area=throat_diameter,
                   pore_diffusivity=pore_diffusivity,
                   throat_diffusivity=throat_diffusivity,
                   conduit_lengths=conduit_lengths,
                   pore_volume=pore_volume,
                   pore_temperature=pore_temperature,
                   throat_temperature=throat_temperature,
                   pore_valence=pore_valence,
                   throat_valence=throat_valence,
                   pore_concentration=pore_concentration,
                   conduit_shape_factors=conduit_shape_factors)


def laplace(target,
            pore_area='pore.area',
            throat_area='throat.area',
            pore_diffusivity='pore.diffusivity',
            throat_diffusivity='throat.diffusivity',
            conduit_lengths='throat.conduit_lengths',
            conduit_shape_factors='throat.poisson_shape_factors',
            pore_volume='pore.volume',
            pore_temperature='pore.temperature',
            throat_temperature='throat.temperature',
            pore_valence='pore.valence',
            throat_valence='throat.valence',
            pore_concentration='pore.concentration'):
    r"""
    Calculate the ionic conductance of conduits in network (using the Laplace
    equation for charge conservation), where a conduit is
    ( 1/2 pore - full throat - 1/2 pore ). See the notes section.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_area : string
        Dictionary key of the pore area values

    throat_area : string
        Dictionary key of the throat area values

    pore_diffusivity : string
        Dictionary key of the pore diffusivity values

    throat_diffusivity : string
        Dictionary key of the throat diffusivity values

    conduit_lengths : string
        Dictionary key of the conduit length values

    conduit_shape_factors : string
        Dictionary key of the conduit DIFFUSION shape factor values

    pore_volume : string
        Dictionary key of the pore volume values

    pore_temperature : string
        Dictionary key of the pore temperature values

    throat_temperature : string
        Dictionary key of the throat temperature values

    pore_valence : string
       Dictionary key of the pore ionic species valence values

    throat_valence : string
       Dictionary key of the throat ionic species valence values

    pore_concentration : string
       Dictionary key of the pore ionic species concentration values

    Returns
    -------
    g : ndarray
        Array containing ionic conductance values for conduits in the
        geometry attached to the given physics object.

    Notes
    -----
    (1) This function requires that all the necessary phase properties already
    be calculated.

    (2) This function calculates the specified property for the *entire*
    network then extracts the values for the appropriate throats at the end.

    (3) This function assumes cylindrical throats with constant cross-section
    area. Corrections for different shapes and variable cross-section area can
    be imposed by passing the proper conduit_shape_factors argument.

    (4) shape_factor depends on the physics of the problem, i.e. diffusion-like
    processes and fluid flow need different shape factors.

    """
    return poisson(target=target,
                   pore_area=pore_area,
                   throat_area=throat_area,
                   pore_diffusivity=pore_diffusivity,
                   throat_diffusivity=throat_diffusivity,
                   conduit_lengths=conduit_lengths,
                   pore_volume=pore_volume,
                   pore_temperature=pore_temperature,
                   throat_temperature=throat_temperature,
                   pore_valence=pore_valence,
                   throat_valence=throat_valence,
                   pore_concentration=pore_concentration,
                   conduit_shape_factors=conduit_shape_factors)


def laplace_2D(target,
               pore_diameter='pore.diameter',
               throat_diameter='throat.diameter',
               pore_diffusivity='pore.diffusivity',
               throat_diffusivity='throat.diffusivity',
               conduit_lengths='throat.conduit_lengths',
               conduit_shape_factors='throat.poisson_shape_factors',
               pore_volume='pore.volume',
               pore_temperature='pore.temperature',
               throat_temperature='throat.temperature',
               pore_valence='pore.valence',
               throat_valence='throat.valence',
               pore_concentration='pore.concentration'):
    r"""
    Calculate the ionic conductance of conduits in network (using the Laplace
    equation for charge conservation), where a conduit is
    ( 1/2 pore - full throat - 1/2 pore ). See the notes section.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_diameter : string
        Dictionary key of the pore diameter values

    throat_diameter : string
        Dictionary key of the throat diameter values

    pore_diffusivity : string
        Dictionary key of the pore diffusivity values

    throat_diffusivity : string
        Dictionary key of the throat diffusivity values

    conduit_lengths : string
        Dictionary key of the conduit length values

    conduit_shape_factors : string
        Dictionary key of the conduit DIFFUSION shape factor values

    pore_volume : string
        Dictionary key of the pore volume values

    pore_temperature : string
        Dictionary key of the pore temperature values

    throat_temperature : string
        Dictionary key of the throat temperature values

    pore_valence : string
       Dictionary key of the pore ionic species valence values

    throat_valence : string
       Dictionary key of the throat ionic species valence values

    pore_concentration : string
       Dictionary key of the pore ionic species concentration values

    Returns
    -------
    g : ndarray
        Array containing ionic conductance values for conduits in the
        geometry attached to the given physics object.

    Notes
    -----
    (1) This function requires that all the necessary phase properties already
    be calculated.

    (2) This function calculates the specified property for the *entire*
    network then extracts the values for the appropriate throats at the end.

    (3) This function assumes cylindrical throats with constant cross-section
    area. Corrections for different shapes and variable cross-section area can
    be imposed by passing the proper conduit_shape_factors argument.

    (4) shape_factor depends on the physics of the problem, i.e. diffusion-like
    processes and fluid flow need different shape factors.

    """
    return laplace(target=target,
                   pore_area=pore_diameter,
                   throat_area=throat_diameter,
                   pore_diffusivity=pore_diffusivity,
                   throat_diffusivity=throat_diffusivity,
                   conduit_lengths=conduit_lengths,
                   conduit_shape_factors=conduit_shape_factors,
                   pore_volume=pore_volume,
                   pore_temperature=pore_temperature,
                   throat_temperature=throat_temperature,
                   pore_valence=pore_valence,
                   throat_valence=throat_valence,
                   pore_concentration=pore_concentration)


def electroneutrality(target,
                      pore_area='pore.area',
                      throat_area='throat.area',
                      pore_diffusivity='pore.diffusivity',
                      throat_diffusivity='throat.diffusivity',
                      conduit_lengths='throat.conduit_lengths',
                      conduit_shape_factors='throat.poisson_shape_factors',
                      pore_volume='pore.volume',
                      pore_temperature='pore.temperature',
                      throat_temperature='throat.temperature',
                      pore_valence='pore.valence',
                      throat_valence='throat.valence',
                      pore_concentration='pore.concentration',
                      ions=[]):
    r"""
    Calculate the ionic conductance of conduits in network (assuming
    electroneutrality for charge conservation), where a conduit is
    ( 1/2 pore - full throat - 1/2 pore ). See the notes section.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_area : string
        Dictionary key of the pore area values

    throat_area : string
        Dictionary key of the throat area values

    pore_diffusivity : string
        Dictionary key of the pore diffusivity values

    throat_diffusivity : string
        Dictionary key of the throat diffusivity values

    conduit_lengths : string
        Dictionary key of the conduit length values

    conduit_shape_factors : string
        Dictionary key of the conduit DIFFUSION shape factor values

    pore_volume : string
        Dictionary key of the pore volume values

    pore_temperature : string
        Dictionary key of the pore temperature values

    throat_temperature : string
        Dictionary key of the throat temperature values

    pore_valence : string
       Dictionary key of the pore ionic species valence values

    throat_valence : string
       Dictionary key of the throat ionic species valence values

    pore_concentration : string
       Dictionary key of the pore ionic species concentration values

    Returns
    -------
    g : ndarray
        Array containing ionic conductance values for conduits in the
        geometry attached to the given physics object.

    Notes
    -----
    (1) This function requires that all the necessary phase properties already
    be calculated.

    (2) This function calculates the specified property for the *entire*
    network then extracts the values for the appropriate throats at the end.

    (3) This function assumes cylindrical throats with constant cross-section
    area. Corrections for different shapes and variable cross-section area can
    be imposed by passing the proper conduit_shape_factors argument.

    (4) shape_factor depends on the physics of the problem, i.e. diffusion-like
    processes and fluid flow need different shape factors.

    """
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    phase = target.project.find_phase(target)
    cn = network['throat.conns'][throats]
    # Getting equivalent areas
    A1 = network[pore_area][cn[:, 0]]
    At = network[throat_area][throats]
    A2 = network[pore_area][cn[:, 1]]
    # Getting conduit lengths
    L1 = network[conduit_lengths + '.pore1'][throats]
    Lt = network[conduit_lengths + '.throat'][throats]
    L2 = network[conduit_lengths + '.pore2'][throats]
    # Preallocating g
    g1, g2, gt = _np.zeros((3, len(Lt)))
    # Setting g to inf when Li = 0 (ex. boundary pores)
    # INFO: This is needed since area could also be zero, which confuses NumPy
    m1, m2, mt = [Li != 0 for Li in [L1, L2, Lt]]
    g1[~m1] = g2[~m2] = gt[~mt] = _sp.inf
    # Getting shape factors
    try:
        SF1 = phase[conduit_shape_factors+'.pore1'][throats]
        SFt = phase[conduit_shape_factors+'.throat'][throats]
        SF2 = phase[conduit_shape_factors+'.pore2'][throats]
    except KeyError:
        SF1 = SF2 = SFt = 1.0
    # Electroneutrality
    F = 96485.3329
    R = 8.3145
    # Getting pores volumes
    Vol1 = network[pore_volume][cn[:, 0]]
    Vol2 = network[pore_volume][cn[:, 1]]
    # Interpolate pore phase property values to throats
    try:
        Tt = phase[throat_temperature][throats]
    except KeyError:
        Tt = phase.interpolate_data(propname=pore_temperature)[throats]
    try:
        T1 = phase[pore_temperature][cn[:, 0]]
        T2 = phase[pore_temperature][cn[:, 1]]
    except KeyError:
        T1 = phase.interpolate_data(propname=throat_temperature)[cn[:, 0]]
        T2 = phase.interpolate_data(propname=throat_temperature)[cn[:, 1]]
    # Iterate over all ions present in the solution
    if ions == []:
        logger.error('List of ions must be provided')
    for i in ions:
        i = '.'+i
        # Check if a concetration field is defined
        try:
            c1 = phase[pore_concentration+i][cn[:, 0]]
            c2 = phase[pore_concentration+i][cn[:, 1]]
        except KeyError:
            c1 = _np.zeros((network.Nt))[cn[:, 0]]
            c2 = _np.zeros((network.Nt))[cn[:, 1]]
        ct = (c1*Vol1 + c2*Vol2)/(Vol1 + Vol2)
        # Interpolate pore phase property values to throats
        try:
            Dt = phase[throat_diffusivity+i][throats]
            Vt = phase[throat_valence+i][throats]
        except KeyError:
            Dt = phase.interpolate_data(
                propname=pore_diffusivity+i)[throats]
            Vt = phase.interpolate_data(
                propname=pore_valence+i)[throats]
        try:
            D1 = phase[pore_diffusivity+i][cn[:, 0]]
            D2 = phase[pore_diffusivity+i][cn[:, 1]]
            V1 = phase[pore_valence+i][cn[:, 0]]
            V2 = phase[pore_valence+i][cn[:, 1]]
        except KeyError:
            D1 = phase.interpolate_data(
                propname=throat_diffusivity+i)[cn[:, 0]]
            D2 = phase.interpolate_data(
                propname=throat_diffusivity+i)[cn[:, 1]]
            V1 = phase.interpolate_data(
                propname=throat_valence+i)[cn[:, 0]]
            V2 = phase.interpolate_data(
                propname=throat_valence+i)[cn[:, 1]]

        g1[m1] += F**2 * V1**2 * (D1*A1*c1)[m1] / (R * T1 * L1[m1])
        g2[m2] += F**2 * V2**2 * (D2*A2*c2)[m1] / (R * T2 * L2[m2])
        gt[mt] += F**2 * Vt**2 * (Dt*At*ct)[mt] / (R * Tt * Lt[mt])
    # Preallocating g_inv
    g_inv1, g_inv2, g_invt = _np.zeros((3, len(Lt)))
    f1, f2, ft = [gi != 0 for gi in [g1, g2, gt]]
    g_inv1[~f1] = g_inv2[~f2] = g_invt[~ft] = _sp.inf
    g_inv1[f1] = 1/g1[f1]
    g_inv2[f2] = 1/g2[f2]
    g_invt[ft] = 1/gt[ft]
    # Apply shape factors and calculate the final conductance
    g = g_inv1/SF1 + g_inv2/SF2 + g_invt/SFt
    g[g != 0] = g[g != 0]**(-1)
    return g


def electroneutrality_2D(target,
                         pore_diameter='pore.diameter',
                         throat_diameter='throat.diameter',
                         pore_diffusivity='pore.diffusivity',
                         throat_diffusivity='throat.diffusivity',
                         conduit_lengths='throat.conduit_lengths',
                         conduit_shape_factors='throat.poisson_shape_factors',
                         pore_volume='pore.volume',
                         pore_temperature='pore.temperature',
                         throat_temperature='throat.temperature',
                         pore_valence='pore.valence',
                         throat_valence='throat.valence',
                         pore_concentration='pore.concentration',
                         ions=[]):
    r"""
    Calculate the ionic conductance of conduits in network (assuming
    electroneutrality for charge conservation), where a conduit is
    ( 1/2 pore - full throat - 1/2 pore ). See the notes section.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_diameter : string
        Dictionary key of the pore diameter values

    throat_diameter : string
        Dictionary key of the throat diameter values

    pore_diffusivity : string
        Dictionary key of the pore diffusivity values

    throat_diffusivity : string
        Dictionary key of the throat diffusivity values

    conduit_lengths : string
        Dictionary key of the conduit length values

    conduit_shape_factors : string
        Dictionary key of the conduit DIFFUSION shape factor values

    pore_volume : string
        Dictionary key of the pore volume values

    pore_temperature : string
        Dictionary key of the pore temperature values

    throat_temperature : string
        Dictionary key of the throat temperature values

    pore_valence : string
       Dictionary key of the pore ionic species valence values

    throat_valence : string
       Dictionary key of the throat ionic species valence values

    pore_concentration : string
       Dictionary key of the pore ionic species concentration values

    Returns
    -------
    g : ndarray
        Array containing ionic conductance values for conduits in the
        geometry attached to the given physics object.

    Notes
    -----
    (1) This function requires that all the necessary phase properties already
    be calculated.

    (2) This function calculates the specified property for the *entire*
    network then extracts the values for the appropriate throats at the end.

    (3) This function assumes cylindrical throats with constant cross-section
    area. Corrections for different shapes and variable cross-section area can
    be imposed by passing the proper conduit_shape_factors argument.

    (4) shape_factor depends on the physics of the problem, i.e. diffusion-like
    processes and fluid flow need different shape factors.

    """
    return electroneutrality(target=target,
                             pore_area=pore_diameter,
                             throat_area=throat_diameter,
                             pore_diffusivity=pore_diffusivity,
                             throat_diffusivity=throat_diffusivity,
                             conduit_lengths=conduit_lengths,
                             conduit_shape_factors=conduit_shape_factors,
                             pore_volume=pore_volume,
                             pore_temperature=pore_temperature,
                             throat_temperature=throat_temperature,
                             pore_valence=pore_valence,
                             throat_valence=throat_valence,
                             pore_concentration=pore_concentration,
                             ions=ions)



r"""

.. autofunction:: openpnm.models.topology.coordination_number

"""
from openpnm.utils import logging
logger = logging.getLogger(__name__)


def coordination_number(target):
    r"""
    """
    network = target.network
    N = network.num_neighbors(pores=network.Ps, flatten=False)
    vals = N[network.pores(target.name)]
    return vals



import numpy
import numpy as np
import scipy as sp
from openpnm.utils import logging
from openpnm.io import GenericIO
from openpnm.network import GenericNetwork
logger = logging.getLogger(__name__)


class PerGeos(GenericIO):
    r"""
    """

    @classmethod
    def save(cls, network=None, phases=[], filename=''):
        r"""
        """
        # avoid printing truncated array
        np.set_printoptions(threshold=np.inf)

        project, network, phases = cls._parse_args(network=network,
                                                   phases=phases)

        # Ensure network has PerGeos' expected properties
        network = network[0]
        if 'pore.EqRadius' not in network.props():
            try:
                network['pore.EqRadius'] = network['pore.diameter']/2
            except KeyError:
                network['pore.EqRadius'] = np.ones([network.Np, ])

        # Add phase properties to network, if any
        for phase in phases:
            for item in phase.keys(mode='props', deep=True):
                temp = item.split('.', 1)
                new_name = temp[0] + '.' + phase.name + '.' + temp[1]
                network[new_name] = phase[item]

        s = ["# Avizo 3D ASCII 3.0\n\n"]
        s.append("define VERTEX " + str(network.Np) + '\n')
        s.append("define EDGE " + str(network.Nt) + '\n')
        s.append("define POINT " + str(2*network.Nt) + '\n\n')
        s.append("Parameters {\n\tContentType \"HxPoreNetworkModel\"\n}\n\n")

        types = {'b': 'int', 'i': 'int', 'f': 'float'}
        typemap = {}
        namemap = {}
        shapemap = {}
        propmap = {}
        i = 1

        NumEdgePoints = 1
        for item in network.keys():
            typemap[item] = types[str(network[item].dtype)[0]]
            ncols = int(network[item].size/network[item].shape[0])
            if ncols > 1:
                shapemap[item] = '[' + str(ncols) + ']'
            else:
                shapemap[item] = ''
            if item.startswith('pore'):
                element = 'pore', 'VERTEX'
            if item.startswith('throat'):
                element = 'throat', 'EDGE'
            n = item.replace(element[0] + '.', '').replace('.', '_').split('_')
            n = ''.join([i[0].upper()+i[1:] for i in n if len(i)])
            namemap[item] = n
            temp = element[1] + " { " + typemap[item] + shapemap[item] + " "\
                   + namemap[item] + " } @" + str(i) + '\n'

            if temp.find('EdgeConnectivity') == -1:
                # replaces openpnm tags with the mandatory am file's tags
                if "Conns" in temp:
                    temp = temp.replace("Conns", "EdgeConnectivity")
                elif "Coords" in temp:
                    temp = temp.replace("Coords", "VertexCoordinates")
                s.append(temp)
            propmap[item] = str(i)
            if "NumEdgePoints" in temp:
                NumEdgePoints = 0
            i += 1

        if NumEdgePoints:
            temp = "EDGE { int NumEdgePoints" + " } @" + str(i) + '\n'
            s.append(temp)
            tempat = "@" + str(i) + '\n'
            i += 1

        # Add POINT data
        s.append("POINT { float[3] EdgePointCoordinates } @" + str(i))
        s.append("\n\n# Data section follows")
        for item in network.keys():
            data = network[item]
            if item != 'throat.EdgeConnectivity':
                s.append('\n\n@' + propmap[item] + '\n')
                if shapemap[item] == '':
                    data = np.atleast_2d(data).T
                if typemap[item] == 'float':
                    formatter = {'float_kind': lambda x: "%.15E" % x}
                else:
                    formatter = None
                if data.dtype == 'bool':
                    data = data.astype(int)
                d = np.array2string(data, formatter=formatter)
                s.append(d.replace('[', '').replace(']', '').replace('\n ', '\n'))

        # Add POINT data
        s.append('\n\n@' + str(i) + '\n')
        formatter = {'float_kind': lambda x: "%.15E" % x}

        conns = network['throat.conns']
        d = np.array2string(network['pore.coords'][conns], formatter=formatter)
        for r in (('[', ''), (']', ''), ('\n\n', '\n'), ('\n  ', '\n'),
                  ('\n ', '\n')):
            d = d.replace(*r)
        d += '\n'
        s.append(d)

        # Add NumEdgePoints
        if NumEdgePoints:
            s.append('\n\n' + tempat)
            s.append(''.join(['2' + '\n']*network.Nt))

        # Write to file
        if filename == '':
            filename = project.name
        fname = cls._parse_filename(filename=filename, ext='am')
        with open(fname, 'w') as f:
            f.write(''.join(s))

    @classmethod
    def load(cls, filename, network=None):
        r"""
        """
        net = {}

        # ---------------------------------------------------------------------
        # Parse the link1 file
        filename = cls._parse_filename(filename=filename, ext='am')
        with open(filename, mode='r') as f:
            Np = None
            Nt = None
            while (Np is None) or (Nt is None):
                s = f.readline()[:-1].split(' ')
                if s[0] == 'define':
                    if s[1] == 'VERTEX':
                        Np = int(s[2])
                    if s[1] == 'EDGE':
                        Nt = int(s[2])

            net = {}
            propmap = {}
            typemap = {}
            shapemap = {}
            while True:
                s = f.readline()[:-1].split(' ')
                if s[0] == 'VERTEX':
                    dshape = [Np]
                    if s[2].endswith(']'):
                        ncols = int(s[2].split('[', 1)[1].split(']')[0])
                        dshape.append(ncols)
                    dtype = s[2].split('[')[0]
                    temp = np.zeros(dshape, dtype=dtype)
                    net['pore.'+s[3]] = temp
                    key = int(s[-1].replace('@', ''))
                    propmap[key] = 'pore.'+s[3]
                    typemap[key] = dtype
                    shapemap[key] = dshape
                elif s[0] == 'EDGE':
                    dshape = [Nt]
                    if s[2].endswith(']'):
                        ncols = int(s[2].split('[', 1)[1].split(']')[0])
                        dshape.append(ncols)
                    dtype = s[2].split('[')[0]
                    temp = np.zeros(dshape, dtype=dtype)
                    net['throat.'+s[3]] = temp
                    key = int(s[-1].replace('@', ''))
                    propmap[key] = 'throat.'+s[3]
                    typemap[key] = dtype
                    shapemap[key] = dshape
                elif s[0] == '#':
                    break

            s = f.read().split('@')
            for key in propmap.keys():
                if key in s:
                    data = s[key].split('\n')[1:]
                    data = ' '.join(data)
                    arr = np.fromstring(data, dtype=typemap[key], sep=' ')
                    arr = np.reshape(arr, newshape=shapemap[key])
                    net[propmap[key]] = arr
            # End file parsing

        net['pore.coords'] = net['pore.VertexCoordinates']
        net['throat.conns'] = np.sort(net['throat.EdgeConnectivity'], axis=1)

        if network is None:
            network = GenericNetwork()
        network = cls._update_network(network=network, net=net)

        return network.project



import os as os
import numpy as np
import scipy as sp
from pathlib import Path
from openpnm.utils import logging
from openpnm.io import GenericIO
from openpnm.network import GenericNetwork
from openpnm.topotools import extend, trim

logger = logging.getLogger(__name__)


class iMorph(GenericIO):
    r"""
    iMorph is a graphical interface program that provides some image analysis
    tools for porous media

    Combines two output files from the iMorph program to build a pore network.
    throats_cellsThroatsGraph_Nodes.txt - stores node shape and type
    information throats_cellsThroatsGraph.txt - stores node connectivity
    """

    @classmethod
    def load(
        cls,
        path,
        node_file="throats_cellsThroatsGraph_Nodes.txt",
        graph_file="throats_cellsThroatsGraph.txt",
        network=None,
        voxel_size=None,
        return_geometry=False,
    ):
        r"""
        Loads network data from an iMorph processed image stack

        Parameters
        ----------
        path : string
            The path of the folder where the subfiles are held

        node_file : string
            The file that describes the pores and throats, the
            default iMorph name is: throats_cellsThroatsGraph_Nodes.txt

        graph_file : string
            The file that describes the connectivity of the network, the
            default iMorph name is: throats_cellsThroatsGraph.txt

        network : OpenPNM Network Object
            The OpenPNM Network onto which the data should be loaded.  If no
            network is supplied then an empty import network is created and
            returned.

        voxel_size : float
            Allows the user to define a voxel size different than what is
            contained in the node_file. The value must be in meters.

        return_geometry : Boolean
            If True, then all geometrical related properties are removed from
            the Network object and added to a GenericGeometry object.  In this
            case the method returns a tuple containing (network, geometry). If
            False (default) then the returned Network will contain all
            properties that were in the original file.  In this case, the user
            can call the ```split_geometry``` method explicitly to perform the
            separation.

        Returns
        -------
        If no Network object is supplied then one will be created and returned.

        If return_geometry is True, then a tuple is returned containing both
        the network and a geometry object.
        """
        #
        path = Path(path)
        node_file = os.path.join(path.resolve(), node_file)
        graph_file = os.path.join(path.resolve(), graph_file)
        # parsing the nodes file
        with open(node_file, "r") as file:
            Np = np.fromstring(file.readline().rsplit("=")[1], sep="\t", dtype=int)[0]
            vox_size = np.fromstring(file.readline().rsplit(")")[1], sep="\t",)[0]

            # network always recreated to prevent errors
            network = GenericNetwork(Np=Np, Nt=0)

            # Define expected properies
            network["pore.volume"] = sp.nan
            scrap_lines = [file.readline() for line in range(4)]
            while True:
                vals = file.readline().split("\t")
                if len(vals) == 1:
                    break
                network["pore.volume"][int(vals[0])] = float(vals[3])
                if "pore." + vals[2] not in network.labels():
                    network["pore." + vals[2]] = False
                network["pore." + vals[2]][int(vals[0])] = True

        if voxel_size is None:
            voxel_size = vox_size * 1.0e-6  # file stores value in microns

        if voxel_size < 0:
            raise (
                Exception(
                    "Error - Voxel size must be specfied in "
                    + "the Nodes file or as a keyword argument."
                )
            )

        # parsing the graph file
        with open(graph_file, "r") as file:
            # Define expected properties
            network["pore.coords"] = np.zeros((Np, 3)) * sp.nan
            network["pore.types"] = sp.nan
            network["pore.color"] = sp.nan
            network["pore.radius"] = sp.nan
            network["pore.dmax"] = sp.nan
            network["pore.node_number"] = sp.nan
            # Scan file to get pore coordinate data
            scrap_lines = [file.readline() for line in range(3)]
            line = file.readline()
            xmax = 0.0
            ymax = 0.0
            zmax = 0.0
            node_num = 0
            while line != "connectivity table\n":
                vals = np.fromstring(line, sep="\t")
                xmax = vals[1] if vals[1] > xmax else xmax
                ymax = vals[2] if vals[2] > ymax else ymax
                zmax = vals[3] if vals[3] > zmax else zmax
                network["pore.coords"][int(vals[0]), :] = vals[1:4]
                network["pore.types"][int(vals[0])] = vals[4]
                network["pore.color"][int(vals[0])] = vals[5]
                network["pore.radius"][int(vals[0])] = vals[6]
                network["pore.dmax"][int(vals[0])] = vals[7]
                network["pore.node_number"][int(vals[0])] = node_num
                node_num += 1
                line = file.readline()
            # Scan file to get to connectivity data
            scrap_lines.append(file.readline())  # Skip line
            # Create sparse lil array incrementally build adjacency matrix
            lil = sp.sparse.lil_matrix((Np, Np), dtype=int)
            while True:
                vals = np.fromstring(file.readline(), sep="\t", dtype=int)
                if len(vals) <= 1:
                    break
                lil.rows[vals[0]] = vals[2:].tolist()
                lil.data[vals[0]] = np.ones(vals[1]).tolist()

        # fixing any negative volumes or distances so they are 1 voxel/micron
        network["pore.volume"][np.where(network["pore.volume"] < 0)[0]] = 1.0
        network["pore.radius"][np.where(network["pore.radius"] < 0)[0]] = 1.0
        network["pore.dmax"][np.where(network["pore.dmax"] < 0)[0]] = 1.0

        # Add adjacency matrix to OpenPNM network
        conns = sp.sparse.triu(lil, k=1, format="coo")
        network.update({"throat.all": np.ones(len(conns.col), dtype=bool)})
        network["throat.conns"] = np.vstack([conns.row, conns.col]).T

        network["pore.to_trim"] = False
        network["pore.to_trim"][network.pores("*throat")] = True
        Ts = network.pores("to_trim")
        new_conns = network.find_neighbor_pores(pores=Ts, flatten=False)
        extend(network=network, throat_conns=new_conns, labels="new_conns")
        for item in network.props("pore"):
            item = item.split(".")[1]
            arr = np.ones_like(network["pore." + item])[0]
            arr = np.tile(A=arr, reps=[network.Nt, 1]) * sp.nan
            network["throat." + item] = np.squeeze(arr)
            network["throat." + item][network.throats("new_conns")] = network[
                "pore." + item
            ][Ts]
        trim(network=network, pores=Ts)

        # setting up boundary pores
        x_coord, y_coord, z_coord = np.hsplit(network["pore.coords"], 3)
        network["pore.front_boundary"] = np.ravel(x_coord == 0)
        network["pore.back_boundary"] = np.ravel(x_coord == xmax)
        network["pore.left_boundary"] = np.ravel(y_coord == 0)
        network["pore.right_boundary"] = np.ravel(y_coord == ymax)
        network["pore.bottom_boundary"] = np.ravel(z_coord == 0)
        network["pore.top_boundary"] = np.ravel(z_coord == zmax)

        # removing any pores that got classified as a boundary pore but
        # weren't labled a border_cell_face
        ps = np.where(
            ~np.in1d(network.pores("*_boundary"), network.pores("border_cell_face"))
        )[0]
        ps = network.pores("*_boundary")[ps]
        for side in ["front", "back", "left", "right", "top", "bottom"]:
            network["pore." + side + "_boundary"][ps] = False
        # setting internal label
        network["pore.internal"] = False
        network["pore.internal"][network.pores("*_boundary", mode="not")] = True

        # adding props to border cell face throats and from pores
        Ts = np.where(
            network["throat.conns"][:, 1] > network.pores("border_cell_face")[0] - 1
        )[0]
        faces = network["throat.conns"][Ts, 1]
        for item in network.props("pore"):
            item = item.split(".")[1]
            network["throat." + item][Ts] = network["pore." + item][faces]
        network["pore.volume"][faces] = 0.0

        # applying unit conversions
        # TODO: Determine if radius and dmax are indeed microns and not voxels
        network["pore.coords"] = network["pore.coords"] * 1e-6
        network["pore.radius"] = network["pore.radius"] * 1e-6
        network["pore.dmax"] = network["pore.dmax"] * 1e-6
        network["pore.volume"] = network["pore.volume"] * voxel_size ** 3
        network["throat.coords"] = network["throat.coords"] * 1e-6
        network["throat.radius"] = network["throat.radius"] * 1e-6
        network["throat.dmax"] = network["throat.dmax"] * 1e-6
        network["throat.volume"] = network["throat.volume"] * voxel_size ** 3

        return network.project



import openpnm
import numpy as np
from openpnm.utils import SettingsDict, logging
logger = logging.getLogger(__name__)


class SettingsDict(SettingsDict):
    def __setitem__(self, key, value):
        if key == 'loglevel':
            logger = logging.getLogger()
            logger.setLevel(value)
        super().__setitem__(key, value)


class Workspace(dict):
    r"""
    The Workspace object provides the highest level of adminstrative
    control over active OpenPNM sessions.

    It is a ``dictionary`` that stores a list of all open Projects by name.

    This class is a
    `singleton <https://en.wikipedia.org/wiki/Singleton_pattern>`_ so that
    whenever and where ever a Workspace is instantiated, the same instance is
    obtained.  This allows it to maintain a definitive record of all open
    Projects.

    See Also
    --------
    Project

    Notes
    -----
    The Workspace object contains a variety of functions that one might expect
    from the 'file-menu' in a typical GUI.  The table below provides a list
    along with a brief description.

    +------------------+-------------------------------------------------+
    | Method           | Description                                     |
    +==================+=================================================+
    | load_workspace   | Loads all saved Projects in a 'pnm' file int... |
    +------------------+-------------------------------------------------+
    | save_workspace   | Saves all the current Projects to a 'pnm' file  |
    +------------------+-------------------------------------------------+
    | new_project      | Creates a new empty Project object              |
    +------------------+-------------------------------------------------+
    | load_project     | Loads a Project from the specified 'pnm' file   |
    +------------------+-------------------------------------------------+
    | save_project     | Saves given Project to a 'pnm' file             |
    +------------------+-------------------------------------------------+
    | close_project    | Removes the specified Project from the Works... |
    +------------------+-------------------------------------------------+
    | copy_project     | Make a copy of an existing Project              |
    +------------------+-------------------------------------------------+

    """

    __instance__ = None

    def __new__(cls, *args, **kwargs):
        if Workspace.__instance__ is None:
            Workspace.__instance__ = dict.__new__(cls)
            cls.settings = SettingsDict()
            cls.settings['loglevel'] = 30
        return Workspace.__instance__

    def __init__(self):
        super().__init__()

    def __setitem__(self, name, project):
        if name is None:
            name = self._gen_name()
        if name in self.keys():
            raise Exception("A project named " + name + " already exists")
        if project in self.values():
            self.pop(project.name, None)
        if not isinstance(project, openpnm.utils.Project):
            project = openpnm.utils.Project(project, name=name)
        super().__setitem__(name, project)

    def copy(self):
        r"""
        """
        raise Exception('Cannot copy Workspace, only one can exist at a time')

    def _create_console_handles(self, project):
        r"""
        Adds all objects in the given project to the console as variables
        with handle names taken from each object's name.
        """
        import __main__
        for item in project:
            __main__.__dict__[item.name] = item

    def save_workspace(self, filename=''):
        r"""
        Saves all the current Projects to a 'pnm' file

        Parameters
        ----------
        filename : string, optional
            If no filename is given, a name is genrated using the current
            time and date. See Notes for more information on valid file names.

        See Also
        --------
        save_project

        Notes
        -----
        The filename can be a string such as 'saved_file.pnm'.  The string can
        include absolute path such as 'C:\networks\saved_file.pnm', or can
        be a relative path such as '..\..\saved_file.pnm', which will look
        2 directories above the current working directory.  It can also be a
        path object object such as that produced by ``pathlib`` or
        ``os.path`` in the Python standard library.

        """
        from openpnm.io import OpenpnmIO
        OpenpnmIO.save_workspace(filename)

    def load_workspace(self, filename, overwrite=False):
        r"""
        Loads a saved Workspace from 'pnm' file into the current Workspace.
        Any Projects present in the current Workspace will be deleted.

        Parameters
        ----------
        filename : string, optional
            The name of the file to open.  See Notes for more information.
        overwrite : boolean
            A flag to indicate if the current Workspace should be
            overwritten when loading the new one.  The default is ``False``,
            meaning the loaded file will be added to the existing data.  Note
            that in this case Project names may clash, in which case the
            newly loaded Projects are given new names.

        See Also
        --------
        load_project

        Notes
        -----
        The filename can be a string such as 'saved_file.pnm'.  The string can
        include absolute path such as 'C:\networks\saved_file.pnm', or can
        be a relative path such as '..\..\saved_file.pnm', which will look
        2 directories above the current working directory.  Can also be a
        path object object such as that produced by ``pathlib`` or
        ``os.path`` in the Python standard library.

        """
        from openpnm.io import OpenpnmIO
        self.clear()
        OpenpnmIO.load_workspace(filename=filename, overwrite=overwrite)

    def save_project(self, project, filename=''):
        r"""
        Saves given Project to a 'pnm' file

        This will include all of associated objects, including algorithms.

        Parameters
        ----------
        project : OpenPNM Project
            The project to save.

        filename : string, optional
            If no filename is given, the given project name is used. See Notes
            for more information.

        See Also
        --------
        save_workspace

        Notes
        -----
        The filename can be a string such as 'saved_file.pnm'.  The string can
        include absolute path such as 'C:\networks\saved_file.pnm', or can
        be a relative path such as '..\..\saved_file.pnm', which will look
        2 directories above the current working directory.  Can also be a
        path object object such as that produced by ``pathlib`` or
        ``os.path`` in the Python standard library.

        """
        from openpnm.io import OpenpnmIO
        OpenpnmIO.save_project(project=project, filename=filename)

    def load_project(self, filename, overwrite=False):
        r"""
        Loads a Project from the specified 'pnm' file

        The loaded project is added to the Workspace . This will *not* delete
        any existing Projects in the Workspace and will rename any Projects
        being loaded if necessary.

        Parameters
        ----------
        filename : string or path object
            The name of the file to open.  See Notes for more information.

        See Also
        --------
        load_workspace

        Notes
        -----
        The filename can be a string such as 'saved_file.pnm'.  The string can
        include absolute path such as 'C:\networks\saved_file.pnm', or can
        be a relative path such as '..\..\saved_file.pnm', which will look
        2 directories above the current working directory.  Can also be a
        path object object such as that produced by ``pathlib`` or
        ``os.path`` in the Python standard library.

        """
        from openpnm.io import OpenpnmIO
        OpenpnmIO.load_project(filename=filename)

    def close_project(self, project):
        r"""
        Removes the specified Project from the Workspace

        This does not save the project, so any changes will be lost.
        """
        del self[project.name]

    def copy_project(self, project, name=None):
        r"""
        Make a copy of an existing Project

        Parameters
        ----------
        project : Project object
            The Project object to be copied

        name : string, optional
            A name for the new copy of the project.  If not supplied, then
            one will be generated (e.g. 'proj_02')

        Returns
        -------
        proj : list
            A handle to the new Project

        """
        proj = project.copy(name)
        return proj

    def new_project(self, name=None):
        r"""
        Creates a new empty Project object

        Parameters
        ----------
        name : string (optional)
            The unique name to give to the project.  If none is given, one
            will be automatically generated (e.g. 'sim_01`)

        Returns
        -------
        proj : list
            An empty Project object, suitable for passing into a Network
            generator

        """
        proj = openpnm.utils.Project(name=name)
        return proj

    def _gen_name(self):
        r"""
        Generates a valid name for projects
        """
        n = [0]
        for item in self.keys():
            if item.startswith('sim_'):
                n.append(int(item.split('sim_')[1]))
        name = 'sim_'+str(max(n)+1).zfill(2)
        return name

    def _gen_ids(self, size):
        r"""
        Generates a sequence of integers of the given ``size``, starting at 1
        greater than the last produced value.

        The Workspace object keeps track of the most recent value, which
        persists until the current python session is restarted, so the
        returned array contains unique values for the given session.

        Parameters
        ----------
        size : int
            The number of values to generate.

        Returns
        -------
        A Numpy array of the specified size, containing integer values starting
        from the last used values.

        Notes
        -----
        When a new Workspace is created the
        """
        if not hasattr(self, '_next_id'):
            # If _next_id has not been set, then assign it
            self._next_id = 0
            # But check ids in any objects present first
            for proj in self.values():
                if len(proj) > 0:
                    if 'pore._id' in proj.network.keys():
                        Pmax = proj.network['pore._id'].max() + 1
                        Tmax = proj.network['throat._id'].max() + 1
                        self._next_id = max([Pmax, Tmax, self._next_id])
        ids = np.arange(self._next_id, self._next_id + size, dtype=np.int64)
        self._next_id += size
        return ids

    def __str__(self):
        s = []
        hr = '―'*78
        s.append(hr)
        s.append('OpenPNM Version ' + openpnm.__version__ + ' Workspace')
        s.append(hr)
        for item in self.values():
            s.append(' ' + item.name)
            s.append(item.__str__())
        return '\n'.join(s)



r"""

.. autofunction:: openpnm.models.geometry.throat_area.cylinder
.. autofunction:: openpnm.models.geometry.throat_area.cuboid

"""
from numpy import pi as _pi


def cylinder(target, throat_diameter='throat.diameter'):
    r"""
    Calculate throat cross-sectional area for a cylindrical throat

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_diameter : string
        Dictionary key of the throat diameter values

    Returns
    -------
    value : NumPy ndarray
        Array containing throat cross-sectional area values.

    """
    diams = target[throat_diameter]
    value = _pi/4*(diams)**2
    return value


def cuboid(target, throat_diameter='throat.diameter'):
    r"""
    Calculate throat cross-sectional area for a cuboid throat

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_diameter : string
        Dictionary key of the throat diameter values

    Returns
    -------
    value : NumPy ndarray
        Array containing throat cross-sectional area values.

    """
    diams = target[throat_diameter]
    value = (diams)**2
    return value


def rectangle(target, throat_diameter='throat.diameter'):
    r"""
    Calculate throat cross-sectional area for a rectangular throat

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_diameter : string
        Dictionary key of the throat diameter values

    """
    return target[throat_diameter]



from flatdict import FlatDict
from openpnm.io import Dict, GenericIO
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class HDF5(GenericIO):
    r"""
    The HDF5 (Hierarchical Data Format) file is good for high-peformance, long
    term data storage
    """

    @classmethod
    def to_hdf5(cls, network=None, phases=[], element=['pore', 'throat'],
                filename='', interleave=True, flatten=False, categorize_by=[]):
        r"""
        Creates an HDF5 file containing data from the specified objects,
        and categorized according to the given arguments.

        Parameters
        ----------
        network : OpenPNM Network Object
            The network containing the desired data

        phases : list of OpenPNM Phase Objects (optional, default is none)
            A list of phase objects whose data are to be included

        element : string or list of strings
            An indication of whether 'pore' and/or 'throat' data are desired.
            The default is both.

        interleave : boolean (default is ``True``)
            When ``True`` (default) the data from all Geometry objects (and
            Physics objects if ``phases`` are given) is interleaved into
            a single array and stored as a network property (or Phase
            property for Physics data). When ``False``, the data for each
            object are stored under their own dictionary key, the structuring
            of which depends on the value of the ``flatten`` argument.

        flatten : boolean (default is ``True``)
            When ``True``, all objects are accessible from the top level
            of the dictionary.  When ``False`` objects are nested under their
            parent object.  If ``interleave`` is ``True`` this argument is
            ignored.

        categorize_by : string or list of strings
            Indicates how the dictionaries should be organized.  The list can
            contain any, all or none of the following strings:

            **'objects'** : If specified the dictionary keys will be stored
            under a general level corresponding to their type (e.g.
            'network/net_01/pore.all'). If  ``interleave`` is ``True`` then
            only the only categories are *network* and *phase*, since
            *geometry* and *physics* data get stored under their respective
            *network* and *phase*.

            **'data'** : If specified the data arrays are additionally
            categorized by ``label`` and ``property`` to separate *boolean*
            from *numeric* data.

            **'elements'** : If specified the data arrays are additionally
            categorized by ``pore`` and ``throat``, meaning that the propnames
            are no longer prepended by a 'pore.' or 'throat.'

        """
        from h5py import File as hdfFile
        project, network, phases = cls._parse_args(network=network,
                                                   phases=phases)
        if filename == '':
            filename = project.name
        filename = cls._parse_filename(filename, ext='hdf')

        dct = Dict.to_dict(network=network, phases=phases, element=element,
                           interleave=interleave, flatten=flatten,
                           categorize_by=categorize_by)
        d = FlatDict(dct, delimiter='/')

        f = hdfFile(filename, "w")
        for item in d.keys():
            tempname = '_'.join(item.split('.'))
            arr = d[item]
            if d[item].dtype == 'O':
                logger.warning(item + ' has dtype object, will not write to file')
                del d[item]
            elif 'U' in str(arr[0].dtype):
                pass
            else:
                f.create_dataset(name='/'+tempname, shape=arr.shape,
                                 dtype=arr.dtype, data=arr)
        return f

    @classmethod
    def from_hdf5(cls):
        r'''
        '''
        raise NotImplementedError()

    def print_levels(f):
        def print_level(f, p='', indent='-'):
            for item in f.keys():
                if hasattr(f[item], 'keys'):
                    p = print_level(f[item], p=p, indent=indent + indent[0])
                elif indent[-1] != ' ':
                    indent = indent + ''
                p = indent + item + '\n' + p
            return(p)
        p = print_level(f)
        print(p)

    def print_flattened(f):
        f.visit(print)



from openpnm.utils import Workspace, logging
from openpnm.physics import GenericPhysics
from openpnm.models import physics as mods
logger = logging.getLogger(__name__)
ws = Workspace()


class Classic(GenericPhysics):
    r"""
    Generic class to generate Physics objects

    Parameters
    ----------
    network : OpenPNM Network object
        The network to which this Physics should be attached

    phase : OpenPNM Phase object
        The Phase object to which this Physics applies

    geometry : OpenPNM Geometry object
        The Geometry object that defines the pores/throats where this Physics
        should be applied.

    name : str, optional
        A unique string name to identify the Physics object, typically same as
        instance name but can be anything.  If left blank, and name will be
        generated that include the class name and a random string.

    """

    def __init__(self, project=None, network=None, phase=None,
                 geometry=None, settings={}, **kwargs):

        # Define some default settings
        self.settings.update({'prefix': 'phys'})
        # Overwrite with user supplied settings, if any
        self.settings.update(settings)

        # Deal with network or project arguments
        if network is not None:
            if project is not None:
                assert network is project.network
            else:
                project = network.project

        super().__init__(project=project, phase=phase, geometry=geometry,
                         **kwargs)

        self.add_model(propname='throat.hydraulic_conductance',
                       model=mods.hydraulic_conductance.classic_hagen_poiseuille)
        self.add_model(propname='throat.diffusive_conductance',
                       model=mods.diffusive_conductance.classic_ordinary_diffusion)
        self.add_model(propname='throat.entry_pressure',
                       model=mods.capillary_pressure.washburn)



import numpy as np
from openpnm.network import Cubic
from openpnm import topotools
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class CubicTemplate(Cubic):
    r"""
    Simple cubic lattice with arbitrary domain shape specified by a template
    image

    The class creates a standard Cubic network the same shape as the provided
    image, then trims pores from the network that are not in the mask.

    Parameters
    ----------
    template : array_like
        The array (image) describing the desired shape of the domain.  All
        locations in the image that are marked as ``True`` are kept while the
        rest of trimmed to yeild the shape.

    spacing : array_like, optional
        The spacing between pore centers in each direction. If not given, then
        [1, 1, 1] is assumed.

    name : string
        An optional name for the object to help identify it.  If not given,
        one will be generated.

    project : OpenPNM Project object, optional
        Each OpenPNM object must be part of a *Project*.  If none is supplied
        then one will be created and this Network will be automatically
        assigned to it.  To create a *Project* use ``openpnm.Project()``.

    Notes
    -----
    The other arguments are the same as ``Cubic`` except that ``shape`` is
    inferred from the ``template`` image.

    See Also
    --------
    The following methods in ``topotools`` can help generate template images:

    template_cylinder_annulus
    template_sphere_shell

    Examples
    --------
    >>> import openpnm as op
    >>> im = op.topotools.template_cylinder_annulus(15, 10, 5)
    >>> pn = op.network.CubicTemplate(template=im)

    And it can be plotted for quick visualization using:

    >>> fig = op.topotools.plot_connections(network=pn)

    .. image:: /../docs/static/images/cubic_template_network.png
        :align: center

    For larger networks and more control over presentation use `Paraview
    <http://www.paraview.org>`_.

    """
    def __init__(self, template, spacing=[1, 1, 1], **kwargs):

        template = np.atleast_3d(template)
        if 'shape' in kwargs:
            del kwargs['shape']
            logger.warning('"shape" argument ignored, inferred from template')
        super().__init__(shape=template.shape, spacing=spacing, **kwargs)

        coords = np.unravel_index(range(template.size), template.shape)
        self['pore.template_coords'] = np.vstack(coords).T
        self['pore.template_indices'] = self.Ps
        topotools.trim(network=self, pores=template.flatten()==0)
        # Add "internal_surface" label to "fake" surface pores!
        ndims = topotools.dimensionality(self).sum()
        max_neighbors = 6 if ndims == 3 else 4
        num_neighbors = np.diff(self.get_adjacency_matrix(fmt="csr").indptr)
        mask_surface = self["pore.surface"]
        mask_internal_surface = (num_neighbors < max_neighbors) & ~mask_surface
        self.set_label("pore.internal_surface", pores=mask_internal_surface)



# -*- coding: utf-8 -*-
"""
===============================================================================
CubicComposite: Generate lattice-like networks with different regions
===============================================================================

"""
import numpy as np
from openpnm.network import GenericNetwork
from openpnm.network import Cubic
from openpnm.topotools import extend, stitch, label_faces

shapes = [[10, 10, 10], [20, 20, 5], [5, 10, 8]]
spacings = [1, 0.5, [2, 1, 0.8]]
origins = [[0, 0, 0], [0, 0, 10], [0, 0, 12.5]]
labels = None


class CubicComposite(GenericNetwork):
    r"""
    """
    def __init__(self, shapes, spacings, origins, labels=None, name=None,
                 project=None):

        # If no labels given, then use default
        if labels is None:
            labels = [None for i in shapes]

        # Generate cubic networks of specified properties and store in a list
        nets = []
        for n in range(len(shapes)):
            net = Cubic(shape=shapes[n], spacing=spacings[n], name=labels[n])
#            label_faces(net)
            net['pore.coords'] += origins[n]
            nets.append(net)

        # Begin the process of merging the networks, but no stitching yet
        main_net = nets.pop(0)
        for net in nets:
            extend(network=main_net, pore_coords=net['pore.coords'],
                   throat_conns=net['throat.conns'] + main_net.Np)

        for net in nets:
            P1 = main_net.pores('surface')
            P1 = net.pores('surface')
            stitch(network=main_net)



import math
import scipy as sp
import numpy as np
from scipy import ndimage
import scipy.spatial as sptl
from openpnm import topotools
from scipy.spatial import ConvexHull
import openpnm.models.geometry as gm
from openpnm.utils import logging, Project
from openpnm.network import DelaunayVoronoiDual
from transforms3d import _gohlketransforms as tr
from openpnm.geometry import GenericGeometry
from openpnm.utils.misc import unique_list

logger = logging.getLogger(__name__)


class VoronoiFibers(Project):
    r"""
    Resembles a fibrous paper or mat with straight intersecting fibers.

    Two geometries are created: DelaunayGeometry defines the pore space
    with pores connected by a Delaunay tesselation and VoronoiGeometry defines
    the fiber space with fibers forming the edges of the Voronoi diagram.
    The two geometries are complimentary and can be accessed individually
    via the project associated with this network object.

    This class is a subclass of the DelaunayVoronoiDual network, and many of
    the parameters behave in the exact same way. However, currently a
    rectangular shape is the only one supported. Additionally, an image is
    created for calculating size data and two parameters relate to this:
    fiber_rad and resolution as detailed below.

    Parameters
    ----------
    num_points : integer
        The number of random base points to distribute inside the domain.
        These points will become connected by the Delaunay triangulation.  The
        points will be generated by calling ``generate_base_points`` in
        ``topotools``.

    points : array_like (num_points x 3)
        A list of coordinates for pre-generated points, typically produced
        using ``generate_base_points`` in topotools.  Note that base points
        should extend beyond the domain so that degenerate Voronoi points
        can be trimmed.  Note: the spherical and cylindrical options
        cannot be used here.

    shape : array_like
        The size and shape of the domain using for generating and trimming
        excess points. It's treated as the outer corner of rectangle [x, y, z]
        whose opposite corner lies at [0, 0, 0].

        By default, a domain size of [1, 1, 1] is used.

    fiber_rad: float
        fiber radius to apply to Voronoi edges when calculating pore and throat
        sizes

    resolution : boolean
        Determines the size of each voxel in the image. Care should be made to
        appropriately set the resolution based on the fiber_radius and the
        shape of the domain so as to remain within memory constraints.

    linear_scale : array_like (len 3)
        This is applied to the domain size before placing the points and then
        reversed afterwards and has the effect of introducing anisotropy into
        an otherwise uniformly distributed point distribution. By default no
        scaling is applied. Applying [1, 1, 2] will stretch the domain by a
        factor of 2 in the z-direction and this will have the affect of
        aligning fibers in the x and y directions once scaling is reversed.

    References
    ----------
    This approach to modeling fibrous materials was first presented by
    Thompson [1] for simulating fluid imbibition in sorbent paper products.
    Gostick [2], and Tranter et al.[3, 4] have subsequently used it to model
    electrodes in fuel cells.

    [1] K. E. Thompson, AlChE J., 48, 1369 (2002)
    [2] J. T. Gostick, Journal of the Electrochemical Society 2013, 160, F731.
    [3] T. G. Tranter et al. Fuel Cells, 2016, 16, 4, 504-515
    [4] T. G. Tranter et al. Transport in Porous Media, 2018, 121, 3, 597-620

    Examples
    --------
    Points will be automatically generated if none are given:

    >>> import openpnm as op
    >>> ws = op.Workspace()
    >>> ws.clear()
    >>> prj = op.materials.VoronoiFibers(num_points=10,
    ...                                  shape=[1e-4, 1e-4, 1e-4],
    ...                                  fiber_rad=5e-6,
    ...                                  resolution=2e-6)
    """

    def __init__(
        self,
        num_points=None,
        points=None,
        shape=[1, 1, 1],
        fiber_rad=None,
        resolution=1e-2,
        name=None,
        linear_scale=None,
        **kwargs
    ):
        super().__init__(name=name)
        shape = np.array(shape)
        scale_applied = False
        if linear_scale is not None:
            if len(linear_scale) != 3:
                logger.exception(
                    msg="linear_scale must have length 3 " + "to scale each axis"
                )
            else:
                ls = np.asarray(linear_scale)
                shape *= ls
                scale_applied = True
        #                net['pore.coords'] *= ls
        if (len(shape) != 3) or np.any(shape == 0):
            raise Exception("Only 3D, rectangular shapes are supported")
        if fiber_rad is None:
            logger.exception(msg="Please initialize class with a fiber_rad")

        net = DelaunayVoronoiDual(
            project=self,
            num_points=num_points,
            points=points,
            shape=shape,
            name=self.name + "_net",
            **kwargs
        )
        if scale_applied:
            net["pore.coords"] /= ls
        net.fiber_rad = fiber_rad
        net.resolution = resolution
        del_geom = DelaunayGeometry(
            project=self,
            network=net,
            pores=net.pores("delaunay"),
            throats=net.throats("delaunay"),
            name=self.name + "_del",
        )

        VoronoiGeometry(
            project=self,
            network=net,
            pores=net.pores("voronoi"),
            throats=net.throats("voronoi"),
            name=self.name + "_vor",
        )
        # Tidy up network
        h = net.check_network_health()
        if len(h["trim_pores"]) > 0:
            topotools.trim(network=net, pores=h["trim_pores"])
        # Copy the Delaunay throat diameters to the boundary pores
        Ps = net.pores(["delaunay", "boundary"], mode="xnor")
        map_Ps = del_geom.map_pores(pores=Ps, origin=net)
        all_Ts = net.find_neighbor_throats(pores=Ps, flatten=False)
        for i, Ts in enumerate(all_Ts):
            Ts = np.asarray(Ts)
            Ts = Ts[net["throat.delaunay"][Ts]]
            if len(Ts) > 0:
                map_Ts = del_geom.map_throats(throats=Ts, origin=net)
                td = del_geom["throat.diameter"][map_Ts]
                ta = del_geom["throat.area"][map_Ts]
                del_geom["pore.diameter"][map_Ps[i]] = td
                del_geom["pore.area"][map_Ps[i]] = ta
        del_geom.regenerate_models(propnames=["throat.equivalent_area"])


class DelaunayGeometry(GenericGeometry):
    r"""
    Subclass of GenericGeometry for the pores connected by throats formed from
    the Delaunay tesselation.

    Parameters
    ----------
    name : string
        A unique name for the network
    """

    def __init__(self, network, **kwargs):
        super().__init__(network=network, **kwargs)
        # Set all the required models
        vertices = network.find_pore_hulls()
        p_coords = np.array(
            [network["pore.coords"][list(p)] for p in vertices], dtype=object
        )
        self["pore.vertices"] = p_coords
        vertices = network.find_throat_facets()
        t_coords = np.array(
            [network["pore.coords"][list(t)] for t in vertices], dtype=object
        )
        self["throat.vertices"] = t_coords

        self.in_hull_volume()
        self["throat.normal"] = self._t_normals()
        self["throat.centroid"] = self._centroids(verts=t_coords)
        self["pore.centroid"] = self._centroids(verts=p_coords)
        (
            self["pore.indiameter"],
            self["pore.incenter"],
        ) = self._indiameter_from_fibers()
        self._throat_props()
        topotools.trim_occluded_throats(network=network, mask=self.name)
        self["throat.volume"] = np.zeros(1, dtype=float)
        self["throat.length"] = np.ones(1, dtype=float)
        self["throat.length"] *= self.network.fiber_rad * 2
        cen_lens = self._throat_c2c()
        self["throat.c2c"] = np.sum(cen_lens, axis=1)
        cen_lens[cen_lens <= 0.0] = 1e-12
        self["throat.conduit_lengths.pore1"] = cen_lens[:, 0]
        self["throat.conduit_lengths.throat"] = cen_lens[:, 1]
        self["throat.conduit_lengths.pore2"] = cen_lens[:, 2]
        # Configurable Models
        self.add_model(
            propname="throat.shape_factor", model=gm.throat_shape_factor.compactness
        )
        self.add_model(propname="pore.diameter", model=gm.pore_size.equivalent_diameter)
        self.add_model(
            propname="pore.area",
            model=gm.pore_area.sphere,
            pore_diameter="pore.diameter",
        )
        self.add_model(
            propname="throat.surface_area", model=gm.throat_surface_area.extrusion
        )
        self.add_model(
            propname="throat.endpoints",
            model=gm.throat_endpoints.straight_throat,
            throat_vector="throat.normal",
        )
        self.regenerate_models()

    def _t_normals(self):
        r"""
        Update the throat normals from the voronoi vertices
        """
        verts = self["throat.vertices"]
        value = np.zeros([len(verts), 3])
        for i in range(len(verts)):
            if len(np.unique(verts[i][:, 0])) == 1:
                verts_2d = np.vstack((verts[i][:, 1], verts[i][:, 2])).T
            elif len(np.unique(verts[i][:, 1])) == 1:
                verts_2d = np.vstack((verts[i][:, 0], verts[i][:, 2])).T
            else:
                verts_2d = np.vstack((verts[i][:, 0], verts[i][:, 1])).T
            hull = sptl.ConvexHull(verts_2d, qhull_options="QJ Pp")
            sorted_verts = verts[i][hull.vertices].astype(float)
            v1 = sorted_verts[-1] - sorted_verts[0]
            v2 = sorted_verts[1] - sorted_verts[0]
            value[i] = tr.unit_vector(np.cross(v1, v2))
        return value

    def _centroids(self, verts):
        r"""
        Function to calculate the centroid as the mean of a set of vertices.
        Used for pore and throat.
        """
        value = np.zeros([len(verts), 3])
        for i, i_verts in enumerate(verts):
            value[i] = np.mean(i_verts, axis=0)
        return value

    def _indiameter_from_fibers(self):
        r"""
        Calculate an indiameter by distance transforming sections of the
        fiber image. By definition the maximum value will be the largest radius
        of an inscribed sphere inside the fibrous hull
        """
        Np = self.num_pores()
        indiam = np.zeros(Np, dtype=float)
        incen = np.zeros([Np, 3], dtype=float)
        hull_pores = np.unique(self._hull_image)
        (Lx, Ly, Lz) = np.shape(self._hull_image)
        (indx, indy, indz) = np.indices([Lx, Ly, Lz])
        indx = indx.flatten()
        indy = indy.flatten()
        indz = indz.flatten()
        for i, pore in enumerate(hull_pores):
            logger.info("Processing pore: " + str(i) + " of " + str(len(hull_pores)))
            dt_pore = self._dt_image * (self._hull_image == pore)
            indiam[pore] = dt_pore.max() * 2
            max_ind = np.argmax(dt_pore)
            incen[pore, 0] = indx[max_ind]
            incen[pore, 1] = indy[max_ind]
            incen[pore, 2] = indz[max_ind]
        indiam *= self.network.resolution
        incen *= self.network.resolution
        return (indiam, incen)

    def _throat_c2c(self):
        r"""
        Calculate the center to center distance from centroid of pore1 to
        centroid of throat to centroid of pore2.
        """
        net = self.network
        Nt = net.num_throats()
        p_cen = net["pore.centroid"]
        t_cen = net["throat.centroid"]
        conns = net["throat.conns"]
        p1 = conns[:, 0]
        p2 = conns[:, 1]
        v1 = t_cen - p_cen[p1]
        v2 = t_cen - p_cen[p2]
        check_nan = ~np.any(np.isnan(v1 + v2), axis=1)
        value = np.ones([Nt, 3], dtype=float) * sp.nan
        for i in range(Nt):
            if check_nan[i]:
                value[i, 0] = sp.linalg.norm(v1[i]) - self.network.fiber_rad
                value[i, 1] = self.network.fiber_rad * 2
                value[i, 2] = sp.linalg.norm(v2[i]) - self.network.fiber_rad
        return value[net.throats(self.name)]

    def _throat_props(self):
        r"""
        Use the Voronoi vertices and perform image analysis to obtain throat
        properties
        """
        from skimage.measure import regionprops
        from skimage.morphology import convex_hull_image

        offset = self.network.fiber_rad
        Nt = self.num_throats()
        centroid = np.zeros([Nt, 3])
        incenter = np.zeros([Nt, 3])
        area = np.zeros(Nt)
        perimeter = np.zeros(Nt)
        inradius = np.zeros(Nt)
        equiv_diameter = np.zeros(Nt)
        eroded_verts = sp.ndarray(Nt, dtype=object)

        res = 200
        vertices = self["throat.vertices"]
        normals = self["throat.normal"]
        z_axis = [0, 0, 1]

        for i in self.throats("delaunay"):
            logger.info("Processing throat " + str(i + 1) + " of " + str(Nt))
            # For boundaries some facets will already be aligned with the axis
            # if this is the case a rotation is unnecessary
            angle = tr.angle_between_vectors(normals[i], z_axis)
            if angle == 0.0 or angle == np.pi:
                # We are already aligned
                rotate_facet = False
                facet = vertices[i]
            else:
                rotate_facet = True
                M = tr.rotation_matrix(
                    tr.angle_between_vectors(normals[i], z_axis),
                    tr.vector_product(normals[i], z_axis),
                )
                facet = np.dot(vertices[i], M[:3, :3].T)
            x = facet[:, 0]
            y = facet[:, 1]
            z = facet[:, 2]
            # Get points in 2d for image analysis
            pts = np.column_stack((x, y))
            # Translate points so min sits at the origin
            translation = [pts[:, 0].min(), pts[:, 1].min()]
            pts -= translation
            order = np.int(math.ceil(-np.log10(np.max(pts))))
            # Normalise and scale the points so that largest span equals the
            # resolution to save on memory and create clear image
            max_factor = np.max([pts[:, 0].max(), pts[:, 1].max()])
            f = res / max_factor
            # Scale the offset and define a structuring element with radius
            r = f * offset
            # Only proceed if r is less than half the span of the image"
            if r <= res / 2:
                pts *= f
                minp1 = pts[:, 0].min()
                minp2 = pts[:, 1].min()
                maxp1 = pts[:, 0].max()
                maxp2 = pts[:, 1].max()
                img = np.zeros(
                    [
                        np.int(math.ceil(maxp1 - minp1) + 1),
                        np.int(math.ceil(maxp2 - minp2) + 1),
                    ]
                )
                int_pts = np.around(pts.astype(float), 0).astype(int)
                for pt in int_pts:
                    img[pt[0]][pt[1]] = 1
                # Pad with zeros all the way around the edges
                img_pad = np.zeros([np.shape(img)[0] + 2, np.shape(img)[1] + 2])
                img_pad[1 : np.shape(img)[0] + 1, 1 : np.shape(img)[1] + 1] = img
                # All points should lie on this plane but could be some
                # rounding errors so use the order parameter
                z_plane = np.unique(np.around(z.astype(float), order + 1))
                if len(z_plane) > 1:
                    logger.error("Throat " + str(i) + " Rotation Failure")
                    temp_arr = np.ones(1)
                    temp_arr.fill(np.mean(z_plane))
                    z_plane = temp_arr
                "Fill in the convex hull polygon"
                convhullimg = convex_hull_image(img_pad)
                # Perform a Distance Transform and black out points less than r
                # to create binary erosion. This is faster than performing an
                # erosion and dt can also be used later to find incircle
                eroded = ndimage.distance_transform_edt(convhullimg)
                eroded[eroded <= r] = 0
                eroded[eroded > r] = 1
                # If we are left with less than 3 non-zero points then the
                # throat is fully occluded
                if np.sum(eroded) >= 3:
                    # Do some image analysis to extract the key properties
                    cropped = eroded[
                        1 : np.shape(img)[0] + 1, 1 : np.shape(img)[1] + 1
                    ].astype(int)
                    regions = regionprops(cropped)
                    # Change this to cope with genuine multi-region throats
                    if len(regions) == 1:
                        for props in regions:
                            x0, y0 = props.centroid
                            equiv_diameter[i] = props.equivalent_diameter
                            area[i] = props.area
                            perimeter[i] = props.perimeter
                            coords = props.coords
                        # Undo the translation, scaling and truncation on the
                        # centroid
                        centroid2d = [x0, y0] / f
                        centroid2d += translation
                        centroid3d = np.concatenate((centroid2d, z_plane))
                        # Distance transform the eroded facet to find the
                        # incenter and inradius
                        dt = ndimage.distance_transform_edt(eroded)
                        temp = np.unravel_index(dt.argmax(), dt.shape)
                        inx0, iny0 = np.asarray(temp).astype(float)
                        incenter2d = [inx0, iny0]
                        # Undo the translation, scaling and truncation on the
                        # incenter
                        incenter2d /= f
                        incenter2d += translation
                        incenter3d = np.concatenate((incenter2d, z_plane))
                        # The offset vertices will be those in the coords that
                        # are closest to the originals
                        offset_verts = []
                        for pt in int_pts:
                            vert = np.argmin(np.sum(np.square(coords - pt), axis=1))
                            if vert not in offset_verts:
                                offset_verts.append(vert)
                        # If we are left with less than 3 different vertices
                        # then the throat is fully occluded as we can't make a
                        # shape with non-zero area
                        if len(offset_verts) >= 3:
                            offset_coords = coords[offset_verts].astype(float)
                            # Undo the translation, scaling and truncation on
                            # the offset_verts
                            offset_coords /= f
                            offset_coords_3d = np.vstack(
                                (
                                    offset_coords[:, 0] + translation[0],
                                    offset_coords[:, 1] + translation[1],
                                    np.ones(len(offset_verts)) * z_plane,
                                )
                            )
                            oc_3d = offset_coords_3d.T
                            # Get matrix to un-rotate the co-ordinates back to
                            # the original orientation if we rotated in the
                            # first place
                            if rotate_facet:
                                MI = tr.inverse_matrix(M)
                                # Unrotate the offset coordinates
                                incenter[i] = np.dot(incenter3d, MI[:3, :3].T)
                                centroid[i] = np.dot(centroid3d, MI[:3, :3].T)
                                eroded_verts[i] = np.dot(oc_3d, MI[:3, :3].T)
                            else:
                                incenter[i] = incenter3d
                                centroid[i] = centroid3d
                                eroded_verts[i] = oc_3d

                            inradius[i] = dt.max()
                            # Undo scaling on other parameters
                            area[i] /= f * f
                            perimeter[i] /= f
                            equiv_diameter[i] /= f
                            inradius[i] /= f
                        else:
                            area[i] = 0
                            perimeter[i] = 0
                            equiv_diameter[i] = 0

        self["throat.area"] = area
        self["throat.perimeter"] = perimeter
        self["throat.centroid"] = centroid
        self["throat.diameter"] = equiv_diameter
        self["throat.indiameter"] = inradius * 2
        self["throat.incenter"] = incenter
        self["throat.offset_vertices"] = eroded_verts

    def inhull(self, xyz, pore, tol=1e-7):
        r"""
        Tests whether points lie within a convex hull or not.
        Computes a tesselation of the hull works out the normals of the facets.
        Then tests whether dot(x.normals) < dot(a.normals) where a is the the
        first vertex of the facets
        """
        xyz = np.around(xyz, 10)
        # Work out range to span over for pore hull
        xmin = xyz[:, 0].min()
        xr = (np.ceil(xyz[:, 0].max()) - np.floor(xmin)).astype(int) + 1
        ymin = xyz[:, 1].min()
        yr = (np.ceil(xyz[:, 1].max()) - np.floor(ymin)).astype(int) + 1
        zmin = xyz[:, 2].min()
        zr = (np.ceil(xyz[:, 2].max()) - np.floor(zmin)).astype(int) + 1

        origin = np.array([xmin, ymin, zmin])
        # start index
        si = np.floor(origin).astype(int)
        xyz -= origin
        dom = np.zeros([xr, yr, zr], dtype=np.uint8)
        indx, indy, indz = np.indices((xr, yr, zr))
        # Calculate the tesselation of the points
        hull = sptl.ConvexHull(xyz)
        # Assume 3d for now
        # Calc normals from the vector cross product of the vectors defined
        # by joining points in the simplices
        vab = xyz[hull.simplices[:, 0]] - xyz[hull.simplices[:, 1]]
        vac = xyz[hull.simplices[:, 0]] - xyz[hull.simplices[:, 2]]
        nrmls = np.cross(vab, vac)
        # Scale normal vectors to unit length
        nrmlen = np.sum(nrmls ** 2, axis=-1) ** (1.0 / 2)
        nrmls = nrmls * np.tile((1 / nrmlen), (3, 1)).T
        # Center of Mass
        center = np.mean(xyz, axis=0)
        # Any point from each simplex
        a = xyz[hull.simplices[:, 0]]
        # Make sure all normals point inwards
        dp = np.sum((np.tile(center, (len(a), 1)) - a) * nrmls, axis=-1)
        k = dp < 0
        nrmls[k] = -nrmls[k]
        # Now we want to test whether dot(x,N) >= dot(a,N)
        aN = np.sum(nrmls * a, axis=-1)
        for plane_index in range(len(a)):
            eqx = nrmls[plane_index][0] * (indx)
            eqy = nrmls[plane_index][1] * (indy)
            eqz = nrmls[plane_index][2] * (indz)
            xN = eqx + eqy + eqz
            dom[xN - aN[plane_index] >= 0 - tol] += 1
        dom[dom < len(a)] = 0
        dom[dom == len(a)] = 1
        ds = np.shape(dom)
        temp_arr = np.zeros_like(self._hull_image, dtype=bool)
        temp_arr[
            si[0] : si[0] + ds[0], si[1] : si[1] + ds[1], si[2] : si[2] + ds[2]
        ] = dom
        self._hull_image[temp_arr] = pore
        del temp_arr

    def in_hull_volume(self):
        r"""
        Work out the voxels inside the convex hull of the voronoi vertices of
        each pore
        """
        i = self.network["pore.internal"]
        s = self.network["pore.surface"]
        d = self.network["pore.delaunay"]
        Ps = self.network.pores()[np.logical_and(d, np.logical_or(i, s))]
        inds = self.network._map(
            ids=self["pore._id"][Ps], element="pore", filtered=True
        )
        # Get the fiber image
        self._get_fiber_image(inds)
        hull_image = np.ones_like(self._fiber_image, dtype=np.uint16) * -1
        self._hull_image = hull_image
        for pore in Ps:
            logger.info("Processing Pore: " + str(pore + 1) + " of " + str(len(Ps)))
            verts = self["pore.vertices"][pore]
            verts = np.asarray(unique_list(np.around(verts, 6)))
            verts /= self.network.resolution
            self.inhull(verts, pore)
        # Catch the voxels that escaped the hulls
        max_h = ndimage.maximum_filter(self._hull_image, size=2)
        mask = self._hull_image == -1
        self._hull_image[mask] = max_h[mask]
        self._process_pore_voxels()

    def _process_pore_voxels(self):
        r"""
        Function to count the number of voxels in the pore and fiber space
        Which are assigned to each hull volume
        """
        from scipy.stats import itemfreq

        num_Ps = self.num_pores()
        pore_vox = np.zeros(num_Ps, dtype=int)
        fiber_vox = np.zeros(num_Ps, dtype=int)
        pore_space = self._hull_image.copy()
        fiber_space = self._hull_image.copy()
        pore_space[self._fiber_image == 0] = -1
        fiber_space[self._fiber_image == 1] = -1
        freq_pore_vox = itemfreq(pore_space)
        freq_pore_vox = freq_pore_vox[freq_pore_vox[:, 0] > -1]
        freq_fiber_vox = itemfreq(fiber_space)
        freq_fiber_vox = freq_fiber_vox[freq_fiber_vox[:, 0] > -1]
        pore_vox[freq_pore_vox[:, 0]] = freq_pore_vox[:, 1]
        fiber_vox[freq_fiber_vox[:, 0]] = freq_fiber_vox[:, 1]
        self["pore.volume"] = pore_vox * self.network.resolution ** 3
        del pore_space
        del fiber_space

    def _bresenham(self, faces, dx):
        r"""
        A Bresenham line function to generate points to fill in for the fibers
        """
        line_points = []
        for face in faces:
            # Get in hull order
            fx = face[:, 0]
            fy = face[:, 1]
            fz = face[:, 2]
            # Find the axis with the smallest spread and remove it to make 2D
            if (np.std(fx) < np.std(fy)) and (np.std(fx) < np.std(fz)):
                f2d = np.vstack((fy, fz)).T
            elif (np.std(fy) < np.std(fx)) and (np.std(fy) < np.std(fz)):
                f2d = np.vstack((fx, fz)).T
            else:
                f2d = np.vstack((fx, fy)).T
            hull = sptl.ConvexHull(f2d, qhull_options="QJ Pp")
            face = np.around(face[hull.vertices].astype(float), 6)
            for i in range(len(face)):
                vec = face[i] - face[i - 1]
                vec_length = np.linalg.norm(vec)
                increments = np.int(np.ceil(vec_length / dx))
                check_p_old = np.array([-1, -1, -1])
                for x in np.linspace(0, 1, increments):
                    check_p_new = face[i - 1] + (vec * x)
                    if np.sum(check_p_new - check_p_old) != 0:
                        line_points.append(check_p_new)
                        check_p_old = check_p_new
        return np.asarray(line_points)

    def _get_fiber_image(self, cpores):
        r"""
        Produce image by filling in voxels along throat edges using Bresenham
        line then performing distance transform on fiber voxels to erode the
        pore space
        """
        fiber_rad = self.network.fiber_rad
        vox_len = self.network.resolution
        # Voxel length of fiber radius
        fiber_rad = np.around((fiber_rad - (vox_len / 2)) / vox_len, 0).astype(int)
        verts = self["throat.vertices"]
        [vxmin, vxmax, vymin, vymax, vzmin, vzmax] = self.vertex_dimension(
            face1=self.pores(), parm="minmax"
        )
        # Translate vertices so that minimum occurs at the origin
        for index in range(len(verts)):
            verts[index] -= np.array([vxmin, vymin, vzmin])
        # Find new size of image array
        cdomain = np.around(
            np.array([(vxmax - vxmin), (vymax - vymin), (vzmax - vzmin)]), 6
        )
        logger.info("Creating fibers in range: " + str(np.around(cdomain, 5)))
        lx = np.int(np.around(cdomain[0] / vox_len) + 1)
        ly = np.int(np.around(cdomain[1] / vox_len) + 1)
        lz = np.int(np.around(cdomain[2] / vox_len) + 1)
        logger.info("Voxels: " + str(lx) + " " + str(ly) + " " + str(lz))
        # Try to create all the arrays we will need at total domain size
        try:
            pore_space = np.ones([lx, ly, lz], dtype=np.uint8)
            fiber_space = np.zeros(shape=[lx, ly, lz], dtype=np.uint8)
            dt = np.zeros([lx, ly, lz], dtype=float)
            # Only need one chunk
            cx = cy = cz = 1
            chunk_len = np.max(np.shape(pore_space))
        except Exception:
            logger.info(
                "Domain too large to fit into memory so chunking "
                + "domain to process image, this may take some time"
            )
            # Do chunking
            chunk_len = 100
            if lx > chunk_len:
                cx = np.ceil(lx / chunk_len).astype(int)
            else:
                cx = 1
            if ly > chunk_len:
                cy = np.ceil(ly / chunk_len).astype(int)
            else:
                cy = 1
            if lz > chunk_len:
                cz = np.ceil(lz / chunk_len).astype(int)
            else:
                cz = 1

        # Get image of the fibers
        line_points = self._bresenham(verts, vox_len / 2)
        line_ints = (np.around((line_points / vox_len), 0)).astype(int)
        for x, y, z in line_ints:
            try:
                pore_space[x][y][z] = 0
            except IndexError:
                logger.warning(
                    "Some elements in image processing are out" + "of bounds"
                )

        num_chunks = np.int(cx * cy * cz)
        cnum = 1
        for ci in range(cx):
            for cj in range(cy):
                for ck in range(cz):
                    # Work out chunk range
                    logger.info(
                        "Processing fiber Chunk: "
                        + str(cnum)
                        + " of "
                        + str(num_chunks)
                    )
                    cxmin = ci * chunk_len
                    cxmax = np.int(np.ceil((ci + 1) * chunk_len + 5 * fiber_rad))
                    cymin = cj * chunk_len
                    cymax = np.int(np.ceil((cj + 1) * chunk_len + 5 * fiber_rad))
                    czmin = ck * chunk_len
                    czmax = np.int(np.ceil((ck + 1) * chunk_len + 5 * fiber_rad))
                    # Don't overshoot
                    if cxmax > lx:
                        cxmax = lx
                    if cymax > ly:
                        cymax = ly
                    if czmax > lz:
                        czmax = lz
                    dt_edt = ndimage.distance_transform_edt
                    dtc = dt_edt(pore_space[cxmin:cxmax, cymin:cymax, czmin:czmax])
                    fiber_space[cxmin:cxmax, cymin:cymax, czmin:czmax][
                        dtc <= fiber_rad
                    ] = 0
                    fiber_space[cxmin:cxmax, cymin:cymax, czmin:czmax][
                        dtc > fiber_rad
                    ] = 1
                    dt[cxmin:cxmax, cymin:cymax, czmin:czmax] = dtc - fiber_rad
                    cnum += 1
        del pore_space
        self._fiber_image = fiber_space
        dt[dt < 0] = 0
        self._dt_image = dt

    def _get_fiber_slice(self, plane=None, index=None):
        r"""
        Plot an image of a slice through the fiber image
        plane contains percentage values of the length of the image in each
        axis

        Parameters
        ----------
        plane : array_like
        List of 3 values, [x,y,z], 2 must be zero and the other must be between
        zero and one representing the fraction of the domain to slice along
        the non-zero axis

        index : array_like
        similar to plane but instead of the fraction an index of the image is
        used
        """
        if hasattr(self, "_fiber_image") is False:
            logger.warning("This method only works when a fiber image exists")
            return None
        if plane is None and index is None:
            logger.warning("Please provide a plane array or index array")
            return None

        if plane is not None:
            if "array" not in plane.__class__.__name__:
                plane = np.asarray(plane)
            if np.sum(plane == 0) != 2:
                logger.warning(
                    "Plane argument must have two zero valued "
                    + "elements to produce a planar slice"
                )
                return None
            t = np.asarray(np.shape(self._fiber_image))
            s = np.around(plane * t).astype(int)
            s[s >= self._fiber_image.shape] -= 1
        elif index is not None:
            if "array" not in index.__class__.__name__:
                index = np.asarray(index)
            if np.sum(index == 0) != 2:
                logger.warning(
                    "Index argument must have two zero valued "
                    + "elements to produce a planar slice"
                )
                return None
            if "int" not in str(index.dtype):
                index = np.around(index).astype(int)
            s = index

        if s[0] != 0:
            slice_image = self._fiber_image[s[0], :, :]
        elif s[1] != 0:
            slice_image = self._fiber_image[:, s[1], :]
        else:
            slice_image = self._fiber_image[:, :, s[2]]

        return slice_image

    def plot_fiber_slice(self, plane=None, index=None, fig=None):
        r"""
        Plot one slice from the fiber image

        Parameters
        ----------
        plane : array_like
        List of 3 values, [x,y,z], 2 must be zero and the other must be between
        zero and one representing the fraction of the domain to slice along
        the non-zero axis

        index : array_like
        similar to plane but instead of the fraction an index of the image is
        used
        """
        import matplotlib.pyplot as plt

        if hasattr(self, "_fiber_image") is False:
            logger.warning("This method only works when a fiber image exists")
            return
        slice_image = self._get_fiber_slice(plane, index)
        if slice_image is not None:
            if fig is None:
                plt.figure()
            plt.imshow(
                slice_image.T, cmap="Greys", origin="lower", interpolation="nearest"
            )

        return fig

    def plot_porosity_profile(self, fig=None):
        r"""
        Return a porosity profile in all orthogonal directions by summing
        the voxel volumes in consectutive slices.
        """
        import matplotlib.pyplot as plt

        if hasattr(self, "_fiber_image") is False:
            logger.warning("This method only works when a fiber image exists")
            return

        im_shape = np.asarray(np.shape(self._fiber_image))
        px = np.zeros(im_shape[0])
        py = np.zeros(im_shape[1])
        pz = np.zeros(im_shape[2])

        for x in np.arange(im_shape[0]):
            px[x] = np.sum(self._fiber_image[x, :, :])
            px[x] /= np.size(self._fiber_image[x, :, :])
        for y in np.arange(im_shape[1]):
            py[y] = np.sum(self._fiber_image[:, y, :])
            py[y] /= np.size(self._fiber_image[:, y, :])
        for z in np.arange(im_shape[2]):
            pz[z] = np.sum(self._fiber_image[:, :, z])
            pz[z] /= np.size(self._fiber_image[:, :, z])

        if fig is None:
            fig = plt.figure()
        ax = fig.gca()
        plots = []
        plots.append(plt.plot(np.arange(im_shape[0]) / im_shape[0], px, "r", label="x"))
        plots.append(plt.plot(np.arange(im_shape[1]) / im_shape[1], py, "g", label="y"))
        plots.append(plt.plot(np.arange(im_shape[2]) / im_shape[2], pz, "b", label="z"))
        plt.xlabel("Normalized Distance")
        plt.ylabel("Porosity")
        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles, labels, loc=1)
        plt.legend(bbox_to_anchor=(1, 1), loc=2, borderaxespad=0.0)
        return fig

    def plot_throat(self, throats, fig=None):
        r"""
        Plot a throat or list of throats in 2D showing key data

        Parameters
        ----------
        throats : list or array containing throat indices tp include in figure

        fig : matplotlib figure object to place plot in
        """
        import matplotlib.pyplot as plt

        throat_list = []
        for throat in throats:
            if throat in range(self.num_throats()):
                throat_list.append(throat)
            else:
                logger.warn("Throat: " + str(throat) + " not part of geometry")
        if len(throat_list) > 0:
            verts = self["throat.vertices"][throat_list]
            offsets = self["throat.offset_vertices"][throat_list]
            normals = self["throat.normal"][throat_list]
            coms = self["throat.centroid"][throat_list]
            incentre = self["throat.incenter"][throat_list]
            inradius = 0.5 * self["throat.indiameter"][throat_list]
            row_col = np.ceil(np.sqrt(len(throat_list)))
            for i in range(len(throat_list)):
                if fig is None:
                    fig = plt.figure()
                ax = fig.add_subplot(row_col, row_col, i + 1)
                vert_2D = self._rotate_and_chop(verts[i], normals[i], [0, 0, 1])
                hull = ConvexHull(vert_2D, qhull_options="QJ Pp")
                for simplex in hull.simplices:
                    plt.plot(
                        vert_2D[simplex, 0], vert_2D[simplex, 1], "k-", linewidth=2
                    )
                plt.scatter(vert_2D[:, 0], vert_2D[:, 1])
                offset_2D = self._rotate_and_chop(offsets[i], normals[i], [0, 0, 1])
                offset_hull = ConvexHull(offset_2D, qhull_options="QJ Pp")
                for simplex in offset_hull.simplices:
                    plt.plot(
                        offset_2D[simplex, 0], offset_2D[simplex, 1], "g-", linewidth=2
                    )
                plt.scatter(offset_2D[:, 0], offset_2D[:, 1])
                # Make sure the plot looks nice by finding the greatest
                # range of points and setting the plot to look square
                xmax = vert_2D[:, 0].max()
                xmin = vert_2D[:, 0].min()
                ymax = vert_2D[:, 1].max()
                ymin = vert_2D[:, 1].min()
                x_range = xmax - xmin
                y_range = ymax - ymin
                if x_range > y_range:
                    my_range = x_range
                else:
                    my_range = y_range
                lower_bound_x = xmin - my_range * 0.5
                upper_bound_x = xmin + my_range * 1.5
                lower_bound_y = ymin - my_range * 0.5
                upper_bound_y = ymin + my_range * 1.5
                plt.axis((lower_bound_x, upper_bound_x, lower_bound_y, upper_bound_y))
                plt.grid(b=True, which="major", color="b", linestyle="-")
                centroid = self._rotate_and_chop(coms[i], normals[i], [0, 0, 1])
                incent = self._rotate_and_chop(incentre[i], normals[i], [0, 0, 1])
                plt.scatter(centroid[0][0], centroid[0][1])
                plt.scatter(incent[0][0], incent[0][1], c="r")
                # Plot incircle
                t = np.linspace(0, 2 * np.pi, 200)
                u = inradius[i] * np.cos(t) + incent[0][0]
                v = inradius[i] * np.sin(t) + incent[0][1]
                plt.plot(u, v, "r-")
                ax.ticklabel_format(style="sci", scilimits=(0, 0))
        else:
            logger.error("Please provide throat indices")
        return fig

    def plot_pore(self, pores, fig=None, axis_bounds=None, include_points=False):
        r"""
        Plot all throats around a given pore or list of pores in 3D

        Parameters
        ----------
        pores : list or array containing pore indices tp include in figure

        fig : matplotlib figure object to place plot in

        axis_bounds : list of [xmin, xmax, ymin, ymax, zmin, zmax] values
            to limit axes to

        include_points : bool
            Determines whether to scatter pore and throat centroids
        """
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d.art3d import Poly3DCollection

        if len(pores) > 0:
            net = self.network
            net_pores = net.map_pores(pores=pores, origin=self)
            centroids = self["pore.centroid"][pores]
            coords = net["pore.coords"][net_pores]
            net_throats = net.find_neighbor_throats(pores=net_pores)
            throats = self.map_throats(throats=net_throats, origin=net)
            tcentroids = self["throat.centroid"][throats]
            # Can't create volume from one throat
            if 1 <= len(throats):
                verts = self["throat.vertices"][throats]
                normals = self["throat.normal"][throats]
                # Get verts in hull order
                ordered_verts = []
                for i in range(len(verts)):
                    vert_2D = self._rotate_and_chop(verts[i], normals[i], [0, 0, 1])
                    hull = ConvexHull(vert_2D, qhull_options="QJ Pp")
                    ordered_verts.append(verts[i][hull.vertices])
                offsets = self["throat.offset_vertices"][throats]
                ordered_offs = []
                for i in range(len(offsets)):
                    offs_2D = self._rotate_and_chop(offsets[i], normals[i], [0, 0, 1])
                    offs_hull = ConvexHull(offs_2D, qhull_options="QJ Pp")
                    ordered_offs.append(offsets[i][offs_hull.vertices])
                # Get domain extents for setting axis
                if axis_bounds is None:
                    [xmin, xmax, ymin, ymax, zmin, zmax] = self.vertex_dimension(
                        net_pores, parm="minmax"
                    )
                else:
                    [xmin, xmax, ymin, ymax, zmin, zmax] = axis_bounds
                if fig is None:
                    fig = plt.figure()
                ax = fig.gca(projection="3d")
                outer_items = Poly3DCollection(
                    ordered_verts, linewidths=1, alpha=0.2, zsort="min"
                )
                outer_face_colours = [(1, 0, 0, 0.01)]
                outer_items.set_facecolor(outer_face_colours)
                ax.add_collection(outer_items)
                inner_items = Poly3DCollection(
                    ordered_offs, linewidths=1, alpha=0.2, zsort="min"
                )
                inner_face_colours = [(0, 0, 1, 0.01)]
                inner_items.set_facecolor(inner_face_colours)
                ax.add_collection(inner_items)
                ax.set_xlim(xmin, xmax)
                ax.set_ylim(ymin, ymax)
                ax.set_zlim(zmin, zmax)
                if include_points:
                    ax.scatter(centroids[:, 0], centroids[:, 1], centroids[:, 2], c="y")
                    ax.scatter(
                        tcentroids[:, 0], tcentroids[:, 1], tcentroids[:, 2], c="r"
                    )
                    ax.scatter(coords[:, 0], coords[:, 1], coords[:, 2], c="b")
                ax.ticklabel_format(style="sci", scilimits=(0, 0))
            else:
                self.plot_throat(throats, fig)
        else:
            logger.error("Please provide pore indices")
        return fig

    def _rotate_and_chop(self, verts, normal, axis=[0, 0, 1]):
        r"""
        Method to rotate a set of vertices (or coords) to align with an axis
        points must be coplanar and normal must be given
        Chops axis coord to give vertices back in 2D
        Used to prepare verts for printing or calculating convex hull in order
        to arrange them in hull order for calculations and printing
        """
        xaxis = [1, 0, 0]
        yaxis = [0, 1, 0]
        zaxis = [0, 0, 1]
        angle = tr.angle_between_vectors(normal, axis)
        if angle == 0.0 or angle == np.pi:
            # We are already aligned
            facet = verts
        else:
            M = tr.rotation_matrix(
                tr.angle_between_vectors(normal, axis), tr.vector_product(normal, axis)
            )
            try:
                facet = np.dot(verts, M[:3, :3].T)
            except ValueError:
                pass
        try:
            x = facet[:, 0]
            y = facet[:, 1]
            z = facet[:, 2]
        except IndexError:
            x = facet[0]
            y = facet[1]
            z = facet[2]
        # Work out span of points and set axes scales to cover this and be
        # equal in both dimensions
        if axis == xaxis:
            output = np.column_stack((y, z))
        elif axis == yaxis:
            output = np.column_stack((x, z))
        elif axis == zaxis:
            output = np.column_stack((x, y))
        else:
            output = facet
        return output

    def vertex_dimension(self, face1=[], face2=[], parm="volume"):
        r"""
        Return the domain extent based on the vertices

        This function is better than using the pore coords as they may be far
        away from the original domain size.  And will alter the effective
        properties which should be based on the original domain sizes. Takes
        one or two sets of pores and works out different geometric properties
        if "length" is specified and two lists are given the planarity is
        determined and the appropriate length (x,y,z) is returned.

        Parameters
        ----------
        face1 : list or array containing pore indices for a face to include in
            calculations.

        parm : string
            Determines what information is returned:
                volume, area (_xy, _xz, _yz), length (_x, _y, _z), minmax.
            Default volume.

        """
        prj = self.project
        network = prj.network
        pores = np.array([], dtype=int)
        if 0 < len(face1):
            pores = np.hstack((pores, face1))
        if 0 < len(face2):
            pores = np.hstack((pores, face2))
        face1_coords = np.around(network["pore.coords"][face1], 12)
        face2_coords = np.around(network["pore.coords"][face2], 12)
        face1_planar = np.zeros(3)
        face2_planar = np.zeros(3)
        planar = np.zeros(3)
        for i in range(3):
            if len(np.unique(face1_coords[:, i])) == 1:
                face1_planar[i] = 1
            if len(np.unique(face2_coords[:, i])) == 1:
                face2_planar[i] = 1
        if 0 < len(face1) and 0 < len(face2):
            planar = face1_planar * face2_planar
        elif 0 < len(face1):
            planar = face1_planar
        elif 0 < len(face2):
            planar = face2_planar
        else:
            return 0
        verts = []
        for pore in pores:
            for vert in np.asarray(list(self["pore.vertices"][pore])):
                verts.append(vert)
        verts = np.asarray(verts)

        vx_min = verts[:, 0].min()
        vx_max = verts[:, 0].max()
        vy_min = verts[:, 1].min()
        vy_max = verts[:, 1].max()
        vz_min = verts[:, 2].min()
        vz_max = verts[:, 2].max()
        output = 0
        width = np.around(vx_max - vx_min, 10)
        depth = np.around(vy_max - vy_min, 10)
        height = np.around(vz_max - vz_min, 10)
        if parm == "volume":
            output = width * depth * height
        elif parm == "area_xy" or (parm == "area" and planar[2] == 1):
            output = width * depth
        elif parm == "area_xz" or (parm == "area" and planar[1] == 1):
            output = width * height
        elif parm == "area_yz" or (parm == "area" and planar[0] == 1):
            output = depth * height
        elif parm == "length_x" or (parm == "length" and planar[0] == 1):
            output = width
        elif parm == "length_y" or (parm == "length" and planar[1] == 1):
            output = depth
        elif parm == "length_z" or (parm == "length" and planar[2] == 1):
            output = height
        elif parm == "minmax":
            output = [vx_min, vx_max, vy_min, vy_max, vz_min, vz_max]
        return output


class VoronoiGeometry(GenericGeometry):
    r"""
    Subclass of GenericGeometry for the fibers making up the complimentary
    solid network.

    Parameters
    ----------
    name : string
        A unique name for the network
    """

    def __init__(self, network, **kwargs):
        super().__init__(network=network, **kwargs)
        rm = "normal"
        net_Ps = network.pores(self.name)
        self["pore.diameter"] = np.ones(self.Np) * network.fiber_rad * 2
        self["pore.indiameter"] = self["pore.diameter"]
        self["pore.incenter"] = network["pore.coords"][net_Ps]
        self["pore.centroid"] = network["pore.coords"][net_Ps]
        self._throat_props()
        self.add_model(
            propname="pore.volume", model=gm.pore_volume.sphere, regen_mode=rm
        )
        self.add_model(propname="pore.area", model=gm.pore_area.sphere, regen_mode=rm)
        self["throat.diameter"] = np.ones(self.Nt) * network.fiber_rad * 2
        self["throat.indiameter"] = self["throat.diameter"]
        self.add_model(
            propname="throat.endpoints",
            model=gm.throat_endpoints.spherical_pores,
            regen_mode=rm,
        )
        self.add_model(
            propname="throat.area", model=gm.throat_area.cylinder, regen_mode=rm
        )
        self.add_model(
            propname="throat.length", model=gm.throat_length.piecewise, regen_mode=rm
        )
        self["throat.c2c"] = self["throat.length"] + network.fiber_rad * 2
        self.add_model(
            propname="throat.volume", model=gm.throat_volume.cylinder, regen_mode=rm
        )
        self.add_model(
            propname="throat.perimeter",
            model=gm.throat_perimeter.cylinder,
            regen_mode=rm,
        )
        self.add_model(
            propname="throat.surface_area",
            model=gm.throat_surface_area.cylinder,
            regen_mode=rm,
        )
        self.add_model(
            propname="throat.shape_factor",
            model=gm.throat_shape_factor.compactness,
            regen_mode=rm,
        )
        self.add_model(
            propname="throat.conduit_lengths", model=gm.throat_length.conduit_lengths
        )

    def _throat_props(self):
        r"""
        Helper Function to calculate the throat normal vectors
        """
        network = self.network
        net_Ts = network.throats(self.name)
        conns = network["throat.conns"][net_Ts]
        p1 = conns[:, 0]
        p2 = conns[:, 1]
        coords = network["pore.coords"]
        normals = tr.unit_vector(coords[p2] - coords[p1])
        self["throat.normal"] = normals
        self["throat.centroid"] = (coords[p1] + coords[p2]) / 2
        self["throat.incenter"] = self["throat.centroid"]



from openpnm.models import misc as _misc


def random(target, seed=None, num_range=[0, 1]):
    return _misc.random(target, element='throat', seed=seed,
                        num_range=num_range)


random.__doc__ = _misc.random.__doc__


def from_neighbor_pores(target, pore_prop='pore.seed', mode='min'):
    return _misc.from_neighbor_pores(target=target, pore_prop=pore_prop,
                                     mode=mode)


random.__doc__ = _misc.from_neighbor_pores.__doc__



import numpy as np
from openpnm.algorithms import ReactiveTransport
from openpnm.models.physics import generic_source_term as gst
from openpnm.utils import logging, Docorator, GenericSettings
logger = logging.getLogger(__name__)
docstr = Docorator()


@docstr.get_sectionsf('IonicConductionSettings', sections=['Parameters'])
@docstr.dedent
class IonicConductionSettings(GenericSettings):
    r"""

    Parameters
    ----------
    phase : OpenPNM Phase object
        The phase on which the algorithm is to be run.
    quantity : str (default = ``'pore.mole_fraction'``)
          The name of the physical quantity to be calculated.
    conductance : str (default = ``'throat.ionic_conductance'``)
        The name of thepore-scale transport conductance values.  These are
        typically calculated by a model attached to a *Physics* object
        associated with the given *Phase*.
    charge_conservation : str (default = ``'electroneutrality'``)
        The assumption adopted to enforce charge conservation when
        performing ions transport simulations.

    Other Parameters
    ----------------

    **The following parameters pertain to the ReactiveTransport class**

    %(ReactiveTransportSettings.other_parameters)s

    ----

    **The following parameters pertain to the GenericTransport class**

    %(GenericTransportSettings.other_parameters)s

    """
    quantity = 'pore.potential'
    conductance = 'throat.ionic_conductance'
    charge_conservation = 'electroneutrality'
    cache_A = False
    cache_b = False


class IonicConduction(ReactiveTransport):
    r"""
    A class to enforce charge conservation in ionic transport simulations.

    Parameters
    ----------
    network : OpenPNM Network object
        The network on which this algorithm operates
    project : OpenPNM Project object
        Either a network or a project must be specified
    name : string, optional
        A unique name to give the object for easier identification.  If not
        given, one is generated.
    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        self.settings._update_settings_and_docs(IonicConductionSettings())
        self.settings.update(settings)

    @docstr.dedent
    def setup(self, phase=None, quantity='', conductance='',
              charge_conservation=None, **kwargs):
        r"""
        This method takes several arguments that are essential to running the
        algorithm and adds them to the settings.

        Parameters
        ----------
        %(IonicConductionSettings.parameters)s

        Notes
        -----
        Any additional arguments are added to the ``settings`` dictionary of
        the object.

        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if conductance:
            self.settings['conductance'] = conductance
        if charge_conservation:
            self.settings['charge_conservation'] = charge_conservation
        super().setup(**kwargs)

    def _charge_conservation_eq_source_term(self, e_alg):
        # Source term for Poisson or charge conservation (electroneutrality) eq
        phase = self.project.phases()[self.settings['phase']]
        Ps = (self['pore.all'] * np.isnan(self['pore.bc_value'])
              * np.isnan(self['pore.bc_rate']))
        mod = gst.charge_conservation
        phys = self.project.find_physics(phase=phase)
        phys[0].add_model(propname='pore.charge_conservation', model=mod,
                          phase=phase, p_alg=self, e_alg=e_alg,
                          assumption=self.settings['charge_conservation'])
        self.set_source(propname='pore.charge_conservation', pores=Ps)



from openpnm.core import Subdomain, ModelsMixin
from openpnm.utils import Workspace, logging
logger = logging.getLogger(__name__)
ws = Workspace()


class GenericGeometry(Subdomain, ModelsMixin):
    r"""
    This generic class is meant as a starter for custom Geometry objects

    It has no pore-scale models assigned to it, so a a blank slate.  Note that
    all OpenPNM Geometry sub-classes are just GenericGeometry instances with a
    number of models added.

    Parameters
    ----------
    network : OpenPNM Network Object
        The Network object to which this Geometry applies.

    pores : array_like
        The list of pores where this Geometry applies.

    throats : array_like
        The list of throats where this Geometry applies.

    name : string
        A unique name to apply to the object.  This name will also be used as a
        label to identify where this Geometry applies.

    project : OpenPNM Project object (optional)
        A Project can be specified instead of ``network``.

    Examples
    --------
    >>> import openpnm as op
    >>> pn = op.network.Cubic(shape=[5, 5, 5])
    >>> Ps = pn.pores('all')  # Get all pores
    >>> Ts = pn.throats('all')  # Get all throats
    >>> geom = op.geometry.GenericGeometry(network=pn, pores=Ps, throats=Ts)

    Now assign pore-scale models to the empty object:

    >>> geom.add_model(propname='pore.size',
    ...                model=op.models.misc.random,
    ...                element='pore',
    ...                num_range=[0.01, 0.1])

    Confirm that the object has one added model:

    >>> print(geom.models)
    ―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    #   Property Name                       Parameter                 Value
    ―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    1   pore.size                           model:                    random
                                            element:                  pore
                                            num_range:                [0.01, 0.1]
                                            seed:                     None
                                            regeneration mode:        normal
    ―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    The results of the model can be seen using the ``show_hist`` function:

    >>> import matplotlib as mpl
    >>> mpl.use('Agg')
    >>> geom.show_hist('pore.size')

    .. image:: /../docs/static/images/generic_geometry_histogram.png
        :width: 500px
        :align: center

    """

    def __init__(self, network=None, project=None, pores=None, throats=None,
                 settings={}, **kwargs):
        # Define some default settings
        self.settings.update({'prefix': 'geo'})
        # Overwrite with user supplied settings, if any
        self.settings.update(settings)

        # Deal with network or project arguments
        if network is not None:
            if project is not None:
                assert network is project.network
            else:
                project = network.project

        super().__init__(project=project, **kwargs)

        network = self.project.network
        if network:
            network['pore.'+self.name] = False
            network['throat.'+self.name] = False
            if (pores is None) and (throats is None):
                logger.info('No pores and throats given, assigning '
                            + self.name + ' to entire domain')
                pores = network.Ps
                throats = network.Ts
            try:
                self._add_locations(pores=pores, throats=throats)
            except Exception as e:
                network.project.purge_object(self)
                logger.error(str(e) +  ', instantiation cancelled')



r"""

.. autofunction:: openpnm.models.physics.meniscus.sinusoidal
.. autofunction:: openpnm.models.physics.meniscus.purcell
.. autofunction:: openpnm.models.physics.meniscus.general_toroidal

"""
import logging
import numpy as np
from openpnm.models.physics.capillary_pressure import _get_key_props
logger = logging.getLogger(__name__)


def purcell(target,
            mode='max',
            target_Pc=None,
            num_points=1e3,
            r_toroid=5e-6,
            throat_diameter='throat.diameter',
            touch_length='throat.touch_length',
            surface_tension='pore.surface_tension',
            contact_angle='pore.contact_angle'):
    r'''
    Wrapper for the general toroidal model to implement the Purcell equation
    for a torus with cylindrical profile and toroidal radius r_toroid.

    Notes
    -----
    This approach accounts for the converging-diverging nature of many throat
    types. Advancing the meniscus beyond the apex of the toroid requires an
    increase in capillary pressure beyond that for a cylindical tube of the
    same radius. The details of this equation are described by Mason and
    Morrow [1]_, and explored by Gostick [2]_ in the context of a pore network
    model.
    '''
    target['throat.scale_a'] = r_toroid
    target['throat.scale_b'] = r_toroid
    output = general_toroidal(target=target,
                              mode=mode,
                              profile_equation='elliptical',
                              target_Pc=target_Pc,
                              num_points=num_points,
                              throat_diameter=throat_diameter,
                              touch_length=touch_length,
                              surface_tension=surface_tension,
                              contact_angle=contact_angle)
    return output


def sinusoidal(target,
               mode='max',
               target_Pc=None,
               num_points=1e3,
               r_toroid=5e-6,
               throat_diameter='throat.diameter',
               pore_diameter='pore.diameter',
               touch_length='throat.touch_length',
               surface_tension='pore.surface_tension',
               contact_angle='pore.contact_angle'):
    r'''
    Wrapper for the general toroidal model to implement the a sinusoidal
    profile. The quarter-wavelength is equal to toroidal radius r_toroid.
    The amplitude is equal to the toroidal radius multiplied by the ratio of
    the throat radius and average connecting pore radius.

    Notes
    -----
    The capillary pressure equation for a sinusoidal throat is extended from
    the Washburn equation as [1]_:

    .. math::
        P_c = -\frac{2\sigma(cos(\alpha + \theta))}{r(x)}

    where alpha is:
    .. math::
        \alpha = arctan(\frac{dr}{dx})

    References
    ----------

    .. [1] A. Forner-Cuenca et. al, Advanced Water Management in PEFCs:
        Diffusion Layers with Patterned Wettability.
        J. ECS. 163, 9, F1038-F1048 (2016).
    '''
    target['throat.scale_a'] = r_toroid
    target['throat.scale_b'] = r_toroid
    output = general_toroidal(target=target,
                              mode=mode,
                              profile_equation='sinusoidal',
                              target_Pc=target_Pc,
                              num_points=num_points,
                              throat_diameter=throat_diameter,
                              touch_length=touch_length,
                              surface_tension=surface_tension,
                              contact_angle=contact_angle)
    return output


def general_toroidal(target,
                     profile_equation='elliptical',
                     mode='max',
                     target_Pc=None,
                     num_points=1e3,
                     throat_scale_a='throat.scale_a',
                     throat_scale_b='throat.scale_b',
                     throat_diameter='throat.diameter',
                     touch_length='throat.touch_length',
                     surface_tension='pore.surface_tension',
                     contact_angle='pore.contact_angle'):
    r'''
    The general model for meniscus properties inside a toroidal throat

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    profile_equation : string (Default is 'elliptical')
        options elliptical, sinusoidal.

    mode : string (Default is 'max')
        Determines what information to send back. Options are:
        'max' : the maximum capillary pressure along the throat axis
        'touch' : the maximum capillary pressure a meniscus can sustain before
                  touching a solid feature
        None : return the meniscus info for a target pressure

    target_Pc : float
        The target capillary pressure to return data for when mode is 'men'

    num_points : float (Default 100)
        The number of divisions to make along the profile length to assess the
        meniscus properties in order to find target pressures, touch lengths,
        minima and maxima.

    throat_scale_a : dict key (string)
        The dictionary key containing the scale factor for adjusting the
        profile along the throat axis (x).

    throat_scale_b : dict key (string)
        The dictionary key containing the scale factor for adjusting the
        profile perpendicular to the throat axis (y).

    throat_diameter : dict key (string)
        The dictionary key containing the throat diameter values to be used.

    touch_length : dict key (string)
        The dictionary key containing the maximum length that a meniscus can
        protrude into the connecting pore before touching a solid feature and
        therfore invading

    surface_tension : dict key (string)
        The dictionary key containing the surface tension values to be used. If
        a pore property is given, it is interpolated to a throat list.

    contact_angle : dict key (string)
        The dictionary key containing the contact angle values to be used. If
        a pore property is given, it is interpolated to a throat list.
    '''
    from sympy import symbols, lambdify
    from sympy import atan as sym_atan
    from sympy import cos as sym_cos
    from sympy import sin as sym_sin
    from sympy import sqrt as sym_sqrt
    from sympy import pi as sym_pi

    # Get data from dictionary keys
    network = target.project.network
    phase = target.project.find_phase(target)
    (element,
     surface_tension,
     contact) = _get_key_props(phase=phase,
                               diameter=throat_diameter,
                               surface_tension=surface_tension,
                               contact_angle=contact_angle)
    # Contact Angle in radians
    contact = np.deg2rad(contact)
    # Network properties
    throatRad = network[throat_diameter]/2
    # Scaling parameters for throat profile
    fa = target[throat_scale_a]
    fb = target[throat_scale_b]
    # Governing equations
    x, a, b, rt, sigma, theta = symbols('x, a, b, rt, sigma, theta')
    if profile_equation == 'elliptical':
        y = sym_sqrt(1 - (x/a)**2)*b
    elif profile_equation == 'sinusoidal':
        y = (sym_cos((sym_pi/2)*(x/a)))*b
    else:
        logger.error('Profile equation is not valid, default to elliptical')
        y = sym_sqrt(1 - (x/a)**2)*b
    # Throat radius profile
    r = rt + (b-y)
    # Derivative of profile
    rprime = r.diff(x)
    # Filling angle
    alpha = sym_atan(rprime)
    # Angle between y axis and contact point to meniscus center
    eta = sym_pi - alpha - theta
    gamma = sym_pi/2 - eta
    # Radius of curvature of meniscus
    rm = r/sym_cos(eta)
    # distance from center of curvature to meniscus contact point (Pythagoras)
    d = rm*sym_sin(eta)
    # angle between throat axis, meniscus center and meniscus contact point
    # Capillary Pressure
    p = 2*sigma/rm
    # Callable functions
    rx = lambdify((x, a, b, rt), r, 'numpy')
    fill_angle = lambdify((x, a, b, rt), alpha, 'numpy')
    rad_curve = lambdify((x, a, b, rt, theta), rm, 'numpy')
    c2x = lambdify((x, a, b, rt, theta), d, 'numpy')
    cap_angle = lambdify((x, a, b, rt, theta), gamma, 'numpy')
    Pc = lambdify((x, a, b, rt, theta, sigma), p, 'numpy')
    # All relative positions along throat
    hp = int(num_points/2)
    log_pos = np.logspace(-4, -1, hp+1)[:-1]
    lin_pos = np.arange(0.1, 1.0, 1/hp)
    half_pos = np.concatenate((log_pos, lin_pos))
    pos = np.concatenate((-half_pos[::-1], half_pos))
    # Now find the positions of the menisci along each throat axis
    Y, X = np.meshgrid(throatRad, pos)
    X *= fa
    # throat Capillary Pressure
    t_Pc = Pc(X, fa, fb, Y, contact, surface_tension)
    # Values of minima and maxima
    Pc_min = np.min(t_Pc, axis=0)
    Pc_max = np.max(t_Pc, axis=0)
    # Arguments of minima and maxima
    a_min = np.argmin(t_Pc, axis=0)
    a_max = np.argmax(t_Pc, axis=0)
    if mode == 'max':
        return Pc_max
    elif mode == 'touch':
        all_rad = rad_curve(X, fa, fb, Y, contact)
        all_c2x = c2x(X, fa, fb, Y, contact)
        all_cen = X - all_c2x
        dist = all_cen + all_rad
        # Only count lengths where meniscus bulges into pore
        dist[all_rad < 0] = 0.0
        touch_len = network[touch_length]
        mask = dist > touch_len
        arg_touch = np.argmax(mask, axis=0)
        # Make sure we only count ones that happen before max pressure
        # And above min pressure (which will be erroneous)
        arg_in_range = (arg_touch < a_max) * (arg_touch > a_min)
        arg_touch[~arg_in_range] = a_max[~arg_in_range]
        x_touch = pos[arg_touch]*fa
        # Return the pressure at which a touch happens
        Pc_touch = Pc(x_touch, fa, fb, throatRad, contact, surface_tension)
        return Pc_touch
    elif target_Pc is None:
        logger.error(msg='Please supply a target capillary pressure'
                     + ' when mode is "men", default to 1.0e-6')
        target_Pc = 1.0e-6
    if np.abs(target_Pc) < 1.0e-6:
        logger.error(msg='Please supply a target capillary pressure'
                     + ' with absolute value greater than 1.0e-6,'
                     + ' default to 1.0e-6')
        target_Pc = 1.0e-6
    # Find the position in-between the minima and maxima corresponding to
    # the target pressure
    inds = np.indices(np.shape(t_Pc))
    # Change values outside the range between minima and maxima to be those
    # Values
    mask = inds[0] < np.ones(len(pos))[:, np.newaxis]*a_min
    t_Pc[mask] = (np.ones(len(pos))[:, np.newaxis]*Pc_min)[mask]
    mask = inds[0] > np.ones(len(pos))[:, np.newaxis]*a_max
    t_Pc[mask] = (np.ones(len(pos))[:, np.newaxis]*Pc_max)[mask]
    # Find the argument at or above the target Pressure
    mask = t_Pc >= target_Pc
    arg_x = np.argmax(mask, axis=0)
    # If outside range change to minima or maxima accordingly
    arg_x[target_Pc < Pc_min] = a_min[target_Pc < Pc_min]
    arg_x[target_Pc > Pc_max] = a_max[target_Pc > Pc_max]
    xpos = pos[arg_x]*fa
    xmin = pos[a_min]*fa
    xmax = pos[a_max]*fa
    # Output
    men_data = {}
    men_data['pos'] = xpos
    men_data['rx'] = rx(xpos, fa, fb, throatRad)
    men_data['alpha'] = fill_angle(xpos, fa, fb, throatRad)
    men_data['alpha_min'] = fill_angle(xmin, fa, fb, throatRad)
    men_data['alpha_max'] = fill_angle(xmax, fa, fb, throatRad)
    men_data['c2x'] = c2x(xpos, fa, fb, throatRad, contact)
    men_data['gamma'] = cap_angle(xpos, fa, fb, throatRad, contact)
    men_data['radius'] = rad_curve(xpos, fa, fb, throatRad, contact)
    # xpos is relative to the throat center
    men_data['center'] = (xpos - men_data['c2x'])
    men_data['men_max'] = men_data['center'] - men_data['radius']

    logger.info(mode+' calculated for Pc: '+str(target_Pc))
    return men_data
